import biom
import skbio
import numpy as np
import pandas as pd
from pandas import concat
from pandas import DataFrame
from typing import Optional
from skbio import OrdinationResults, DistanceMatrix, TreeNode
from gemelli.factorization import TensorFactorization
from gemelli.rpca import rpca_table_processing
from gemelli.preprocessing import (build,
                                   fast_unifrac,
                                   bp_read_phylogeny,
                                   retrieve_t2t_taxonomy,
                                   create_taxonomy_metadata)
from gemelli._defaults import (DEFAULT_COMP, DEFAULT_MSC,
                               DEFAULT_MFC, DEFAULT_BL,
                               DEFAULT_MTD, DEFAULT_MFF,
                               DEFAULT_TENSALS_MAXITER,
                               DEFAULT_FMETA as DEFFM)
# import QIIME2 if in a Q2env otherwise set type to str
try:
    from q2_types.tree import NewickFormat
except ImportError:
    # python does not check but technically this is the type
    NewickFormat = str


def phylogenetic_ctf_without_taxonomy(
                     table: biom.Table,
                     phylogeny: NewickFormat,
                     sample_metadata: DataFrame,
                     individual_id_column: str,
                     state_column: str,
                     n_components: int = DEFAULT_COMP,
                     min_sample_count: int = DEFAULT_MSC,
                     min_feature_count: int = DEFAULT_MFC,
                     min_feature_frequency: float = DEFAULT_MFF,
                     min_depth: int = DEFAULT_MTD,
                     max_iterations_als: int = DEFAULT_TENSALS_MAXITER,
                     max_iterations_rptm: int = DEFAULT_TENSALS_MAXITER,
                     n_initializations: int = DEFAULT_TENSALS_MAXITER) -> (
                         OrdinationResults, OrdinationResults,
                         DistanceMatrix, DataFrame, DataFrame,
                         TreeNode, biom.Table, biom.Table):
    """
    Runs phylogenetic CTF without taxonomy. This code will
    be run QIIME2 versions of gemelli. Outside of QIIME2
    please use phylogenetic_ctf.
    """
    # run CTF helper and parse output for QIIME
    output = phylogenetic_ctf(table=table,
                              phylogeny=phylogeny,
                              sample_metadata=sample_metadata,
                              individual_id_column=individual_id_column,
                              state_column=state_column,
                              n_components=n_components,
                              min_sample_count=min_sample_count,
                              min_feature_count=min_feature_count,
                              min_feature_frequency=min_feature_frequency,
                              min_depth=min_depth,
                              max_iterations_als=max_iterations_als,
                              max_iterations_rptm=max_iterations_rptm,
                              n_initializations=n_initializations)
    (ord_res, state_ordn, dists, straj, ftraj,
     phylogeny, counts_by_node,
     _, subject_table) = output

    return (ord_res, state_ordn, dists, straj, ftraj,
            phylogeny, counts_by_node, subject_table)


def phylogenetic_ctf_with_taxonomy(
                     table: biom.Table,
                     phylogeny: NewickFormat,
                     taxonomy: pd.DataFrame,
                     sample_metadata: DataFrame,
                     individual_id_column: str,
                     state_column: str,
                     n_components: int = DEFAULT_COMP,
                     min_sample_count: int = DEFAULT_MSC,
                     min_feature_count: int = DEFAULT_MFC,
                     min_feature_frequency: float = DEFAULT_MFF,
                     min_depth: int = DEFAULT_MTD,
                     max_iterations_als: int = DEFAULT_TENSALS_MAXITER,
                     max_iterations_rptm: int = DEFAULT_TENSALS_MAXITER,
                     n_initializations: int = DEFAULT_TENSALS_MAXITER) -> (
                         OrdinationResults, OrdinationResults,
                         DistanceMatrix, DataFrame, DataFrame,
                         TreeNode, biom.Table,
                         pd.DataFrame, biom.Table):
    """
    Runs phylogenetic CTF with taxonomy. This code will
    be run QIIME2 versions of gemelli. Outside of QIIME2
    please use phylogenetic_ctf.
    """
    # feature_metadata
    taxonomy = taxonomy.to_dataframe()
    # run CTF helper and parse output for QIIME
    output = phylogenetic_ctf(table=table,
                              phylogeny=phylogeny,
                              taxonomy=taxonomy,
                              sample_metadata=sample_metadata,
                              individual_id_column=individual_id_column,
                              state_column=state_column,
                              n_components=n_components,
                              min_sample_count=min_sample_count,
                              min_feature_count=min_feature_count,
                              min_feature_frequency=min_feature_frequency,
                              min_depth=min_depth,
                              max_iterations_als=max_iterations_als,
                              max_iterations_rptm=max_iterations_rptm,
                              n_initializations=n_initializations)
    (ord_res, state_ordn, dists, straj, ftraj,
     phylogeny, counts_by_node,
     result_taxonomy, subject_table) = output

    return (ord_res, state_ordn,
            dists, straj, ftraj,
            phylogeny, counts_by_node,
            result_taxonomy, subject_table)


def phylogenetic_ctf(table: biom.Table,
                     phylogeny: NewickFormat,
                     sample_metadata: DataFrame,
                     individual_id_column: str,
                     state_column: str,
                     taxonomy: Optional[pd.DataFrame] = None,
                     n_components: int = DEFAULT_COMP,
                     min_sample_count: int = DEFAULT_MSC,
                     min_feature_count: int = DEFAULT_MFC,
                     min_feature_frequency: float = DEFAULT_MFF,
                     min_depth: int = DEFAULT_MTD,
                     max_iterations_als: int = DEFAULT_TENSALS_MAXITER,
                     max_iterations_rptm: int = DEFAULT_TENSALS_MAXITER,
                     n_initializations: int = DEFAULT_TENSALS_MAXITER) -> (
                         OrdinationResults, OrdinationResults,
                         DistanceMatrix, DataFrame, DataFrame,
                         TreeNode, biom.Table,
                         Optional[pd.DataFrame], biom.Table):
    """
    Phylogenetic Compositional Tensor Factorization (CTF) with
    mode 3 tensor. This means subjects have repeated measures
    across only one axis (e.g. time or space).
    For more information see (1-4).

    Parameters
    ----------
    table: numpy.ndarray, required
    The feature table in biom format containing the
    samples over which metric should be computed.

    phylogeny: str, required
    Path to the file containing the phylogenetic tree containing tip
    identifiers that correspond to the feature identifiers in the table.
    This tree can contain tip ids that are not present in the table,
    but all feature ids in the table must be present in this tree.

    sample_metadata: DataFrame, required
    Sample metadata file in QIIME2 formatting. The file must
    contain the columns for individual_id_column and
    state_column and the rows matched to the table.

    individual_id_column: str, required
    Metadata column containing subject IDs to use for
    pairing samples. WARNING: if replicates exist for an
    individual ID at either state_1 to state_N, that
    subject will be mean grouped by default.

    state_column: str, required
    Metadata column containing state (e.g.,Time,
    BodySite) across which samples are paired. At least
    one is required but up to four are allowed by other
    state inputs.

    taxonomy: pd.DataFrame, optional
    Taxonomy file in QIIME2 formatting. See the feature metdata
    section of https://docs.qiime2.org/2021.11/tutorials/metadata

    n_components: int, optional : Default is 3
    The underlying rank of the data and number of
    output dimentions.

    min_sample_count: int, optional : Default is 0
    Minimum sum cutoff of sample across all features.
    The value can be at minimum zero and must be an
    whole integer. It is suggested to be greater than
    or equal to 500.

    min_feature_count: int, optional : Default is 0
    Minimum sum cutoff of features across all samples.
    The value can be at minimum zero and must be
    an whole integer.

    min_feature_frequency: float, optional : Default is 0
    Minimum percentage of samples a feature must appear
    with a value greater than zero. This value can range
    from 0 to 100 with decimal values allowed.

    max_als_iterations: int, optional
    Max number of Alternating Least Square (ALS).

    tol_als: float, optional
    The minimization -- convergence break point for ALS.

    max_rtpm_iterations: int, optional
    Max number of Robust Tensor Power Method (RTPM) iterations.

    n_initializations: int, optional
    The number of initial vectors. Larger values will
    give more accurate factorization but will be more
    computationally expensive.

    Returns
    -------
    OrdinationResults
        Compositional biplot of subjects as points and
        features as arrows. Where the variation between
        subject groupings is explained by the log-ratio
        between opposing arrows. WARNING: The % variance
        explained is only spread over n-components and
        can be inflated.

    OrdinationResults
        Compositional biplot of states as points and
        features as arrows. Where the variation between
        subject groupings is explained by the log-ratio
        between opposing arrows. WARNING: The % variance
        explained is only spread over n-components and can be
        inflated.

    DistanceMatrix
        A sample-sample distance matrix generated from the
        euclidean distance of the subject-state ordinations
        and itself.

    DataFrame
        A trajectory is an ordination that can be
        visualizedover time or another context.

    DataFrame
        A trajectory is an ordination that can be
        visualizedover time or another context.

    Raises
    ------
    ValueError
        `ValueError: n_components must be at least 2`.

    ValueError
        `ValueError: Data-table contains either np.inf or -np.inf`.

    ValueError
        `ValueError: The n_components must be less
            than the minimum shape of the input tensor`.

    References
    ----------
    .. [1] Martino C, Shenhav L, Marotz CA, Armstrong G, McDonald D,
           Vázquez-Baeza Y, Morton JT, Jiang L, Dominguez-Bello MG,
           Swafford AD, Halperin E, Knight R. 2020.
           Context-aware dimensionality reduction deconvolutes
           gut microbial community dynamics.
           Nat Biotechnol https://doi.org/10.1038/s41587-020-0660-7.
    .. [2] Jain, Prateek, and Sewoong Oh. 2014.
            “Provable Tensor Factorization with Missing Data.”
            In Advances in Neural Information Processing Systems
            27, edited by Z. Ghahramani, M. Welling, C. Cortes,
            N. D. Lawrence, and K. Q. Weinberger, 1431–39.
            Curran Associates, Inc.
    .. [3] A. Anandkumar, R. Ge, M., Janzamin,
            Guaranteed Non-Orthogonal Tensor
            Decomposition via Alternating Rank-1
            Updates. CoRR (2014),
            pp. 1-36.
    .. [4] A. Anandkumar, R. Ge, D. Hsu,
            S. M. Kakade, M. Telgarsky,
            Tensor Decompositions for Learning Latent Variable Models
            (A Survey for ALT). Lecture Notes in Computer Science
            (2015), pp. 19–38.

    Examples
    --------
    import numpy as np
    import pandas as pd
    from biom import Table
    from gemelli.ctf import phylogenetic_ctf

    # make a table
    X = np.array([[9, 3, 0, 0],
                [9, 9, 0, 1],
                [0, 1, 4, 5],
                [0, 0, 3, 4],
                [1, 0, 8, 9]])
    sample_ids = ['s1','s2','s3','s4']
    feature_ids = ['f1','f2','f3','f4','f5']
    bt = Table(X, feature_ids, sample_ids)
    # make mock metadata
    mf = pd.DataFrame([[i//2, i%2] for i, s in enumerate(sample_ids)],
                    sample_ids, ['subject_id', 'context'])
    # write an example tree to read
    f = open("demo-tree.nwk", "w")
    newick = '(((f1:1,f2:1)n9:1,f3:1)n8:1,(f4:1,f5:1)n2:1)n1:1;'
    f.write(newick)
    f.close()
    # make mock taxonomy
    taxonomy = pd.DataFrame({fid:['k__kingdom; p__phylum;'
                                'c__class; o__order; '
                                'f__family; g__genus;'
                                's__',
                                0.99]
                            for fid in feature_ids},
                            ['Taxon', 'Confidence']).T

    # run phylo-CTF with taxonomy
    # subject 1 will seperate from subject 2
    (subject_biplot, state_biplot,
    distance_matrix,
    state_subject_ordination,
    state_feature_ordination,
    counts_by_node_tree,
    counts_by_node,
    _,
    subject_table) = phylogenetic_ctf(bt, 'demo-tree.nwk', mf,
                                    'subject_id', 'context')


    # run phylo-CTF without taxonomy
    # subject 1 will seperate from subject 2
    (subject_biplot, state_biplot,
    distance_matrix,
    state_subject_ordination,
    state_feature_ordination,
    counts_by_node_tree,
    counts_by_node,
    t2t_taxonomy,
    subject_table) = phylogenetic_ctf(bt, 'demo-tree.nwk', mf,
                                    'subject_id', 'context', taxonomy)

    """

    # run CTF helper and parse output for QIIME
    helper_results = phylogenetic_ctf_helper(table,
                                             phylogeny,
                                             sample_metadata,
                                             individual_id_column,
                                             [state_column],
                                             n_components,
                                             min_sample_count,
                                             min_feature_count,
                                             min_feature_frequency,
                                             min_depth,
                                             max_iterations_als,
                                             max_iterations_rptm,
                                             n_initializations,
                                             taxonomy)
    (state_ordn, ord_res, dists, straj,
     ftraj, phylogeny, counts_by_node,
     result_taxonomy, subject_table) = helper_results

    # save only first state (QIIME can't handle a list yet)
    dists = list(dists.values())[0]
    straj = list(straj.values())[0]
    ftraj = list(ftraj.values())[0]
    state_ordn = list(state_ordn.values())[0]

    return (ord_res, state_ordn,
            dists, straj, ftraj,
            phylogeny, counts_by_node,
            result_taxonomy, subject_table)


def phylogenetic_ctf_helper(table: biom.Table,
                            phylogeny: NewickFormat,
                            sample_metadata: DataFrame,
                            individual_id_column: str,
                            state_column: list,
                            n_components: int = DEFAULT_COMP,
                            min_sample_count: int = DEFAULT_MSC,
                            min_feature_count: int = DEFAULT_MFC,
                            min_feature_frequency: float = DEFAULT_MFF,
                            min_depth: int = DEFAULT_MTD,
                            max_iterations_als: int = DEFAULT_TENSALS_MAXITER,
                            max_iterations_rptm: int = DEFAULT_TENSALS_MAXITER,
                            n_initializations: int = DEFAULT_TENSALS_MAXITER,
                            taxonomy: Optional[pd.DataFrame] = None) -> (
                                OrdinationResults, OrdinationResults,
                                DistanceMatrix, DataFrame, DataFrame,
                                TreeNode, biom.Table, biom.Table):
    """Helper function. Please use phylogenetic_ctf directly."""
    # check the table for validity and then filter
    process_results = ctf_table_processing(table,
                                           sample_metadata,
                                           individual_id_column,
                                           state_column,
                                           min_sample_count,
                                           min_feature_count,
                                           min_feature_frequency,
                                           taxonomy)
    (table, sample_metadata,
     all_sample_metadata, taxonomy) = process_results
    # import the tree
    phylogeny = bp_read_phylogeny(table, phylogeny, min_depth)
    # build the vectorized table
    counts_by_node, tree_index, branch_lengths, fids, otu_ids\
        = fast_unifrac(table, phylogeny)
    # import expanded table
    counts_by_node = biom.Table(counts_by_node.T,
                                fids, table.ids())
    # make a table for Empress
    subject_table = per_subject_table(table,
                                      sample_metadata.copy(),
                                      individual_id_column)
    # use taxonomy if provided
    result_taxonomy = None
    if taxonomy is not None:
        # collect taxonomic information for all tree nodes.
        traversed_taxonomy = retrieve_t2t_taxonomy(phylogeny, taxonomy)
        result_taxonomy = create_taxonomy_metadata(phylogeny,
                                                   traversed_taxonomy)
    # build the tensor object and factor - return results
    tensal_results = tensals_helper(counts_by_node,
                                    sample_metadata,
                                    all_sample_metadata,
                                    individual_id_column,
                                    state_column,
                                    branch_lengths,
                                    n_components,
                                    max_iterations_als,
                                    max_iterations_rptm,
                                    n_initializations,
                                    result_taxonomy)
    state_ordn, ord_res, dists, straj, ftraj = tensal_results

    return (state_ordn, ord_res,
            dists, straj, ftraj,
            phylogeny, counts_by_node,
            result_taxonomy, subject_table)


def ctf(table: biom.Table,
        sample_metadata: DataFrame,
        individual_id_column: str,
        state_column: str,
        n_components: int = DEFAULT_COMP,
        min_sample_count: int = DEFAULT_MSC,
        min_feature_count: int = DEFAULT_MFC,
        min_feature_frequency: float = DEFAULT_MFF,
        max_iterations_als: int = DEFAULT_TENSALS_MAXITER,
        max_iterations_rptm: int = DEFAULT_TENSALS_MAXITER,
        n_initializations: int = DEFAULT_TENSALS_MAXITER,
        feature_metadata: DataFrame = DEFFM) -> (OrdinationResults,
                                                 OrdinationResults,
                                                 DistanceMatrix,
                                                 DataFrame,
                                                 DataFrame):
    """
    Compositional Tensor Factorization (CTF) with mode 3 tensor. This
    means subjects have repeated measures across only one axis
    (e.g. time or space). For more information see (1-4).

    Parameters
    ----------
    table: numpy.ndarray, required
    The feature table in biom format containing the
    samples over which metric should be computed.

    sample_metadata: DataFrame, required
    Sample metadata file in QIIME2 formatting. The file must
    contain the columns for individual_id_column and
    state_column and the rows matched to the table.

    individual_id_column: str, required
    Metadata column containing subject IDs to use for
    pairing samples. WARNING: if replicates exist for an
    individual ID at either state_1 to state_N, that
    subject will be mean grouped by default.

    state_column: str, required
    Metadata column containing state (e.g.,Time,
    BodySite) across which samples are paired. At least
    one is required but up to four are allowed by other
    state inputs.

    n_components: int, optional : Default is 3
    The underlying rank of the data and number of
    output dimentions.

    min_sample_count: int, optional : Default is 0
    Minimum sum cutoff of sample across all features.
    The value can be at minimum zero and must be an
    whole integer. It is suggested to be greater than
    or equal to 500.

    min_feature_count: int, optional : Default is 0
    Minimum sum cutoff of features across all samples.
    The value can be at minimum zero and must be
    an whole integer.

    min_feature_frequency: float, optional : Default is 0
    Minimum percentage of samples a feature must appear
    with a value greater than zero. This value can range
    from 0 to 100 with decimal values allowed.

    max_als_iterations: int, optional
    Max number of Alternating Least Square (ALS).

    tol_als: float, optional
    The minimization -- convergence break point for ALS.

    max_rtpm_iterations: int, optional
    Max number of Robust Tensor Power Method (RTPM) iterations.

    n_initializations: int, optional
    The number of initial vectors. Larger values will
    give more accurate factorization but will be more
    computationally expensive.

    feature_metadata: pd.DataFrame, optional
    Taxonomy file in QIIME2 formatting. See the feature metdata
    section of https://docs.qiime2.org/2021.11/tutorials/metadata

    Returns
    -------
    OrdinationResults
        Compositional biplot of subjects as points and
        features as arrows. Where the variation between
        subject groupings is explained by the log-ratio
        between opposing arrows. WARNING: The % variance
        explained is only spread over n-components and
        can be inflated.

    OrdinationResults
        Compositional biplot of states as points and
        features as arrows. Where the variation between
        subject groupings is explained by the log-ratio
        between opposing arrows. WARNING: The % variance
        explained is only spread over n-components and can be
        inflated.

    DistanceMatrix
        A sample-sample distance matrix generated from the
        euclidean distance of the subject-state ordinations
        and itself.

    DataFrame
        A trajectory is an ordination that can be
        visualizedover time or another context.

    DataFrame
        A trajectory is an ordination that can be
        visualizedover time or another context.

    Raises
    ------
    ValueError
        `ValueError: n_components must be at least 2`.

    ValueError
        `ValueError: Data-table contains either np.inf or -np.inf`.

    ValueError
        `ValueError: The n_components must be less
            than the minimum shape of the input tensor`.

    References
    ----------
    .. [1] Martino C, Shenhav L, Marotz CA, Armstrong G, McDonald D,
           Vázquez-Baeza Y, Morton JT, Jiang L, Dominguez-Bello MG,
           Swafford AD, Halperin E, Knight R. 2020.
           Context-aware dimensionality reduction deconvolutes
           gut microbial community dynamics.
           Nat Biotechnol https://doi.org/10.1038/s41587-020-0660-7.
    .. [2] Jain, Prateek, and Sewoong Oh. 2014.
            “Provable Tensor Factorization with Missing Data.”
            In Advances in Neural Information Processing Systems
            27, edited by Z. Ghahramani, M. Welling, C. Cortes,
            N. D. Lawrence, and K. Q. Weinberger, 1431–39.
            Curran Associates, Inc.
    .. [3] A. Anandkumar, R. Ge, M., Janzamin,
            Guaranteed Non-Orthogonal Tensor
            Decomposition via Alternating Rank-1
            Updates. CoRR (2014),
            pp. 1-36.
    .. [4] A. Anandkumar, R. Ge, D. Hsu,
            S. M. Kakade, M. Telgarsky,
            Tensor Decompositions for Learning Latent Variable Models
            (A Survey for ALT). Lecture Notes in Computer Science
            (2015), pp. 19–38.

    Examples
    --------
    import numpy as np
    import pandas as pd
    from biom import Table
    from gemelli.ctf import ctf

    # make a table
    X = np.array([[9, 3, 0, 0],
                [9, 9, 0, 1],
                [0, 1, 4, 5],
                [0, 0, 3, 4],
                [1, 0, 8, 9]])
    sample_ids = ['s1','s2','s3','s4']
    feature_ids = ['f1','f2','f3','f4','f5']
    bt = Table(X, feature_ids, sample_ids)
    # make mock metadata
    mf = pd.DataFrame([[i//2, i%2] for i, s in enumerate(sample_ids)],
                    sample_ids, ['subject_id', 'context'])
    # run CTF
    # subject 1 will seperate from subject 2
    (subject_biplot, state_biplot,
    distance_matrix,
    state_subject_ordination,
    state_feature_ordination) = ctf(bt, mf, 'subject_id', 'context')

    """
    # run CTF helper and parse output for QIIME
    helper_results = ctf_helper(table,
                                sample_metadata,
                                individual_id_column,
                                [state_column],
                                n_components,
                                min_sample_count,
                                min_feature_count,
                                min_feature_frequency,
                                max_iterations_als,
                                max_iterations_rptm,
                                n_initializations,
                                feature_metadata)
    state_ordn, ord_res, dists, straj, ftraj = helper_results
    # save only first state (QIIME can't handle a list yet)
    dists = list(dists.values())[0]
    straj = list(straj.values())[0]
    ftraj = list(ftraj.values())[0]
    state_ordn = list(state_ordn.values())[0]

    return ord_res, state_ordn, dists, straj, ftraj


def ctf_helper(table: biom.Table,
               sample_metadata: DataFrame,
               individual_id_column: str,
               state_column: list,
               n_components: int = DEFAULT_COMP,
               min_sample_count: int = DEFAULT_MSC,
               min_feature_count: int = DEFAULT_MFC,
               min_feature_frequency: float = DEFAULT_MFF,
               max_iterations_als: int = DEFAULT_TENSALS_MAXITER,
               max_iterations_rptm: int = DEFAULT_TENSALS_MAXITER,
               n_initializations: int = DEFAULT_TENSALS_MAXITER,
               feature_metadata: DataFrame = DEFFM) -> (
                   OrdinationResults, OrdinationResults,
                   DistanceMatrix, DataFrame, DataFrame):
    """Helper function. Please use ctf directly."""
    # check the table for validity and then filter
    process_results = ctf_table_processing(table,
                                           sample_metadata,
                                           individual_id_column,
                                           state_column,
                                           min_sample_count,
                                           min_feature_count,
                                           min_feature_frequency,
                                           feature_metadata)
    (table, sample_metadata,
     all_sample_metadata, feature_metadata) = process_results
    # build the tensor object and factor - return results
    tensal_results = tensals_helper(table,
                                    sample_metadata,
                                    all_sample_metadata,
                                    individual_id_column,
                                    state_column,
                                    None,
                                    n_components,
                                    max_iterations_als,
                                    max_iterations_rptm,
                                    n_initializations,
                                    feature_metadata)
    state_ordn, ord_res, dists, straj, ftraj = tensal_results

    return state_ordn, ord_res, dists, straj, ftraj


def per_subject_table(table: biom.Table,
                      sample_metadata: DataFrame,
                      individual_id_column: str):
    """ builds a per-subject summed table for Empress
    """

    # convert biom.Table into dataframe
    subject_table = DataFrame(table.matrix_data.toarray(),
                              table.ids('observation'),
                              table.ids('sample')).T
    # get subject ID information
    subject_sample_mf = sample_metadata.copy()[individual_id_column]
    subject_sample_mf = subject_sample_mf.reindex(subject_table.index)
    subject_table[individual_id_column] = subject_sample_mf
    # sum by subject across all samples
    subject_table = subject_table.groupby(individual_id_column).sum().T
    # back to biom.Table
    subject_table = biom.Table(subject_table.values,
                               subject_table.index.astype(str),
                               subject_table.columns.astype(str))

    return subject_table


def ctf_table_processing(table: biom.Table,
                         sample_metadata: DataFrame,
                         individual_id_column: str,
                         state_columns: list,
                         min_sample_count: int = DEFAULT_MSC,
                         min_feature_count: int = DEFAULT_MFC,
                         min_feature_frequency: float = DEFAULT_MFF,
                         feature_metadata: DataFrame = DEFFM) -> (
                             dict, OrdinationResults, dict, tuple):
    """ Runs  Compositional Tensor Factorization CTF.
    """

    # validate the metadata using q2 as a wrapper
    if sample_metadata is not None and not isinstance(sample_metadata,
                                                      DataFrame):
        sample_metadata = sample_metadata.to_dataframe()
    keep_cols = state_columns + [individual_id_column]
    all_sample_metadata = sample_metadata.drop(keep_cols, axis=1)
    sample_metadata = sample_metadata[keep_cols]
    # validate the metadata using q2 as a wrapper
    if feature_metadata is not None and not isinstance(feature_metadata,
                                                       DataFrame):
        feature_metadata = feature_metadata.to_dataframe()
    # match the data (borrowed in part from gneiss.util.match)
    subtablefids = table.ids('observation')
    subtablesids = table.ids('sample')
    if len(subtablesids) != len(set(subtablesids)):
        raise ValueError('Data-table contains duplicate sample IDs')
    if len(subtablefids) != len(set(subtablefids)):
        raise ValueError('Data-table contains duplicate feature IDs')
    submetadataids = set(sample_metadata.index)
    subtablesids = set(subtablesids)
    subtablefids = set(subtablefids)
    if feature_metadata is not None:
        submetadatafeat = set(feature_metadata.index)
        fidx = subtablefids & submetadatafeat
        if len(fidx) == 0:
            raise ValueError(("No more features left.  Check to make "
                              "sure that the sample names between "
                              "`feature-metadata` and `table` are "
                              "consistent"))
        feature_metadata = feature_metadata.reindex(fidx)
    sidx = subtablesids & submetadataids
    if len(sidx) == 0:
        raise ValueError(("No more features left.  Check to make sure that "
                          "the sample names between `sample-metadata` and"
                          " `table` are consistent"))
    if feature_metadata is not None:
        table.filter(list(fidx), axis='observation', inplace=True)
    table.filter(list(sidx), axis='sample', inplace=True)
    sample_metadata = sample_metadata.reindex(sidx)

    # filter the table (same as RPCA)
    table = rpca_table_processing(table,
                                  min_sample_count,
                                  min_feature_count,
                                  min_feature_frequency)

    # return data based on input
    if feature_metadata is not None:
        return (table, sample_metadata,
                all_sample_metadata,
                feature_metadata)
    else:
        return (table, sample_metadata,
                all_sample_metadata, None)


def tensals_helper(table: biom.Table,
                   sample_metadata: DataFrame,
                   all_sample_metadata: DataFrame,
                   individual_id_column: str,
                   state_columns: list,
                   branch_lengths: np.array = DEFAULT_BL,
                   n_components: int = DEFAULT_COMP,
                   max_iterations_als: int = DEFAULT_TENSALS_MAXITER,
                   max_iterations_rptm: int = DEFAULT_TENSALS_MAXITER,
                   n_initializations: int = DEFAULT_TENSALS_MAXITER,
                   feature_metadata: DataFrame = DEFFM) -> (
                       dict, OrdinationResults, dict, tuple):
    """ Runs  Compositional Tensor Factorization CTF.
    """

    # table to dataframe
    table = DataFrame(table.matrix_data.toarray(),
                      table.ids('observation'),
                      table.ids('sample'))

    # tensor building
    tensor = build()
    tensor.construct(table, sample_metadata,
                     individual_id_column,
                     state_columns,
                     branch_lengths=branch_lengths)

    # factorize
    TF = TensorFactorization(
        n_components=n_components,
        max_als_iterations=max_iterations_als,
        max_rtpm_iterations=max_iterations_rptm,
        n_initializations=n_initializations).fit(tensor.rclr_transformed)

    # label tensor loadings
    TF.label(tensor, taxonomy=feature_metadata)

    # if the n_components is two add PC3 of zeros
    # this is referenced as in issue in
    # <https://github.com/biocore/emperor/commit
    # /a93f029548c421cb0ba365b4294f7a5a6b0209ce>
    if n_components == 2:
        TF.subjects.loc[:, 'PC3'] = [0] * len(TF.subjects.index)
        TF.features.loc[:, 'PC3'] = [0] * len(TF.features.index)
        TF.proportion_explained['PC3'] = 0
        TF.eigvals['PC3'] = 0

    # save ordination results
    short_method_name = 'CTF_Biplot'
    long_method_name = 'Compositional Tensor Factorization Biplot'
    # only keep PC -- other tools merge metadata
    keep_PC = [col for col in TF.features.columns if 'PC' in col]
    subj_ordin = OrdinationResults(
        short_method_name,
        long_method_name,
        TF.eigvals,
        samples=TF.subjects[keep_PC].dropna(axis=0),
        features=TF.features[keep_PC].dropna(axis=0),
        proportion_explained=TF.proportion_explained)
    # save distance matrix for each condition
    distances = {}
    state_ordn = {}
    subject_trajectories = {}
    feature_trajectories = {}
    for condition, cond, dist, straj, ftraj in zip(tensor.conditions,
                                                   TF.conditions,
                                                   TF.subject_distances,
                                                   TF.subject_trajectory,
                                                   TF.feature_trajectory):
        # match distances to metadata
        ids = straj.index
        ind_dict = dict((ind, ind_i) for ind_i, ind in enumerate(ids))
        inter = set(ind_dict).intersection(sample_metadata.index)
        indices = sorted([ind_dict[ind] for ind in inter])
        dist = dist[indices, :][:, indices]
        distances[condition] = skbio.stats.distance.DistanceMatrix(
            dist, ids=ids[indices])
        # fix conditions
        if n_components == 2:
            cond['PC3'] = [0] * len(cond.index)
        cond = OrdinationResults(short_method_name,
                                 long_method_name,
                                 TF.eigvals,
                                 samples=cond[keep_PC].dropna(axis=0),
                                 features=TF.features[keep_PC].dropna(axis=0),
                                 proportion_explained=TF.proportion_explained)
        state_ordn[condition] = cond
        # add the sample metadata before returning output
        # additionally only keep metadata with trajectory
        # output available.
        pre_merge_cols = list(straj.columns)
        straj = concat([straj.reindex(all_sample_metadata.index),
                        all_sample_metadata],
                       axis=1, sort=True)
        straj = straj.dropna(subset=pre_merge_cols)
        # ensure index name for q2
        straj.index.name = "#SampleID"
        # save traj.
        keep_PC_traj = [col for col in straj.columns
                        if 'PC' in col]
        straj[keep_PC_traj] -= straj[keep_PC_traj].mean()
        ftraj[keep_PC_traj] -= ftraj[keep_PC_traj].mean()
        subject_trajectories[condition] = straj
        ftraj.index = ftraj.index.astype(str)
        feature_trajectories[condition] = ftraj
    return (state_ordn, subj_ordin, distances,
            subject_trajectories, feature_trajectories)



import unittest
import numpy as np
from scipy.linalg import qr
from gemelli.factorization import TensorFactorization, khatri_rao


class TestTenAls(unittest.TestCase):

    def setUp(self):
        # generate random noiseless low-rank orthogonal tensor
        r = 3  # rank is 2
        self.r = r
        n1 = 10
        n2 = 12
        n3 = 8
        n4 = 6
        n5 = 2
        U01 = np.random.rand(n1, r)
        U02 = np.random.rand(n2, r)
        U03 = np.random.rand(n3, r)
        U04 = np.random.rand(n4, r)
        U05 = np.random.rand(n5, r)

        # QR factorization ensures component factors are orthogonal
        U1, temp = qr(U01)
        U2, temp = qr(U02)
        U3, temp = qr(U03)
        U4, temp = qr(U04)
        U5 = U05
        U1 = U1[:, 0:r]
        U2 = U2[:, 0:r]
        U3 = U3[:, 0:r]
        U4 = U4[:, 0:r]
        U5 = U5[:, 0:r]
        T = np.zeros((n1, n2, n3))
        for i in range(n3):
            T[:, :, i] = np.matmul(U1, np.matmul(np.diag(U3[i, :]), U2.T))
        to_multiply = [U5, U5, U4, U5]
        to_multiply5 = [U5, U5, U4, U5, U5]
        product = khatri_rao(to_multiply)
        product5 = khatri_rao(to_multiply5)
        T4 = product.sum(1).reshape((n5, n5, n4, n5))
        T5 = product5.sum(1).reshape((n5, n5, n4, n5, n5))
        # sample entries
        p = 2 * (r ** 0.5 * np.log(n1 * n2 * n3)) / np.sqrt(n1 * n2 * n3)
        p4 = 1 * (r ** 5e-5 * np.log(n5 * n5 * n4 * n5)) / np.sqrt(n5 * n5 * n4
                                                                   * n5)
        p5 = 1 * (r ** 5e-5 * np.log(n5 * n5 * n4 * n5 * n5)) / \
            np.sqrt(n5 * n5 * n4 * n5 * n5)
        self.E = abs(np.ceil(np.random.rand(n1, n2, n3) - 1 + p))
        self.E4 = abs(np.ceil(np.random.rand(n5, n5, n4, n5) - 1 + p4))
        self.E5 = abs(np.ceil(np.random.rand(n5, n5, n4, n5, n5) - 1 + p5))
        self.TE = T * self.E
        self.TE4 = T4 * self.E4
        self.TE5 = T5 * self.E5
        self.TE_noise = self.TE + (0.0001 / np.sqrt(n1 * n2 * n3)
                                   * np.random.randn(n1, n2, n3) * self.E)
        self.TE_noise4 = self.TE4 + (1e-10 / np.sqrt(n5 * n5 * n4 * n5)
                                     * np.random.randn(n5, n5,
                                                       n4, n5) * self.E4)
        self.TE_noise5 = self.TE5 + (1e-10 / np.sqrt(n5 * n5 * n4 * n5 * n5)
                                     * np.random.randn(n5, n5, n4, n5, n5
                                                       ) * self.E5)
        self.n1 = n1
        self.n2 = n2
        self.n3 = n3
        self.n4 = n4
        self.n5 = n5
        self.U1 = U1
        self.U2 = U2
        self.U3 = U3
        self.U4 = U4
        self.U5 = U5

    def test_TenAls_center(self):
        # TensorFactorization no noise w/ centered
        TF = TensorFactorization(center=True).fit(self.TE)
        L1, L2, L3 = TF.loadings
        for l1v in L1.mean(axis=1):
            self.assertLess(l1v, 1)
        for l2v in L2.mean(axis=1):
            self.assertLess(l2v, 1)

    def test_TenAls_noiseless(self):
        # TensorFactorization no noise
        TF = TensorFactorization(center=False).fit(self.TE)
        L1, L2, L3 = TF.loadings
        s = TF.eigvals
        # test accuracy
        rmse = 0
        for i3 in range(self.n3):
            A1 = self.U1
            A2 = np.matmul(self.U2, np.diag(self.U3[i3, :]))
            B1 = L1
            B2 = np.matmul(L2, np.diag(L3[i3, :] * s.T.flatten()))
            rmse += np.trace(np.matmul(np.matmul(A1.T, A1), np.matmul(A2.T,
                                                                      A2))) + \
                np.trace(np.matmul(np.matmul(B1.T, B1), np.matmul(B2.T,
                                                                  B2))) + \
                -2 * np.trace(np.matmul(np.matmul(B1.T, A1), np.matmul(A2.T,
                                                                       B2)))
        self.assertTrue(1e2 > abs(rmse))

    def test_TenAls_mode4_noiseless(self):
        # TODO check values
        TF = TensorFactorization(center=False).fit(self.TE4)
        L1, L2, L3, L4 = TF.loadings
        # test accuracy

    def test_TenAls_mode5_noiseless(self):
        # TODO check values
        TF = TensorFactorization(center=False).fit(self.TE5)
        L1, L2, L3, L4, L5 = TF.loadings

    def test_TenAls_noise(self):
        # TensorFactorization no noise
        TF = TensorFactorization(center=False).fit(self.TE_noise)
        L1, L2, L3 = TF.loadings
        s = TF.eigvals
        # test accuracy
        rmse = 0
        for i3 in range(self.n3):
            A1 = self.U1
            A2 = np.matmul(self.U2, np.diag(self.U3[i3, :]))
            B1 = L1
            B2 = np.matmul(L2, np.diag(L3[i3, :] * s.T.flatten()))
            rmse += np.trace(np.matmul(np.matmul(A1.T, A1), np.matmul(A2.T,
                                                                      A2))) + \
                np.trace(np.matmul(np.matmul(B1.T, B1), np.matmul(B2.T,
                                                                  B2))) + \
                -2 * np.trace(np.matmul(np.matmul(B1.T, A1), np.matmul(A2.T,
                                                                       B2)))
        self.assertTrue(1e2 > abs(rmse))

    def test_TenAls_mode4_noise(self):
        # TODO check values
        TF = TensorFactorization(center=False).fit(self.TE_noise4)
        L1, L2, L3, L4 = TF.loadings

    def test_TenAls_mode5_noise(self):
        # TODO check values
        TF = TensorFactorization(center=False).fit(self.TE_noise5)
        L1, L2, L3, L4, L5 = TF.loadings

    def test_khatri_rao(self):
        multiply_2 = khatri_rao([self.U1, self.U2])
        self.assertEqual((self.n1 * self.n2, self.r), multiply_2.shape)
        multiply_3 = khatri_rao([self.U1, self.U2, self.U3])
        self.assertEqual((self.n1 * self.n2 * self.n3, self.r),
                         multiply_3.shape)
        multiply_4 = khatri_rao([self.U1, self.U2, self.U3, self.U4])
        self.assertEqual((self.n1 * self.n2 * self.n3 * self.n4, self.r),
                         multiply_4.shape)
        multiply_5 = khatri_rao([self.U1, self.U2, self.U3, self.U4, self.U4])
        self.assertEqual((self.n1 * self.n2 * self.n3 * self.n4 * self.n4,
                          self.r),
                         multiply_5.shape)
        multiply_6 = khatri_rao([self.U1, self.U2, self.U3, self.U4,
                                 self.U4, self.U5])
        self.assertEqual((self.n1 * self.n2 * self.n3 * self.n4 * self.n4 *
                          self.n5, self.r),
                         multiply_6.shape)

    def test_errors(self):
        # test not array
        with self.assertRaises(ValueError):
            TensorFactorization(center=False).fit(list(range(10)))
        # test if none missing
        with self.assertRaises(ValueError):
            TensorFactorization(center=False).fit(np.ones((5, 5)))
        # test no nan(s)
        TE_errors = self.TE
        TE_errors[0, :, :] = np.inf
        with self.assertRaises(ValueError):
            TensorFactorization(center=False).fit(TE_errors)
        # test max rank
        with self.assertRaises(ValueError):
            paeram_ = np.max(self.TE_noise.shape) + 10
            TensorFactorization(center=False,
                                n_components=paeram_).fit(self.TE_noise)



from qiime2.plugin import SemanticType
from q2_types.sample_data import SampleData
from q2_types.feature_data import FeatureData

SampleTrajectory = SemanticType(
    'SampleTrajectory', variant_of=SampleData.field['type'])
FeatureTrajectory = SemanticType(
    'FeatureTrajectory', variant_of=FeatureData.field['type'])



# ----------------------------------------------------------------------------
# Copyright (c) 2019--, gemelli development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# ----------------------------------------------------------------------------

import warnings
import numpy as np
import pandas as pd
from scipy.linalg import svd
from numpy.linalg import norm
from .base import _BaseImpute
from scipy.spatial import distance
from gemelli.optspace import rank_estimate


class TensorFactorization(_BaseImpute):

    def __init__(self,
                 n_components=3,
                 max_als_iterations=50,
                 tol_als=1e-7,
                 max_rtpm_iterations=50,
                 n_initializations=50,
                 tol_rtpm=1e-5,
                 fillna=1.0,
                 center=True,
                 check_dense=True):
        """

        This class performs a low-rank N-order
        tensor factorization for partially observered
        non-symmetric sets [1]. This method relies on a
        CANDECOMP/PARAFAC (CP) tensor decomposition [2,3].
        Missing values are handled by an alternating
        least squares (ALS) minimization between
        tensor and the reconstructed tensor.

        Parameters
        ----------
        n_components : int, optional
            The underlying low-rank, will be
            equal to the number of rank 1
            components that are output. The
            higher the rank given, the more
            expensive the computation will
            be.
        max_als_iterations : int, optional
            Max number of Alternating Least Square (ALS).
        tol_als : float, optional
            The minimization -- convergence break point for
            ALS.
        max_rtpm_iterations : int, optional
            Max number of Robust Tensor Power Method (RTPM)
            iterations.
        n_initializations : int, optional
            The number of initial
            vectors. Larger values will
            give more accurate factorization
            but will be more computationally
            expensive.
        tol_rtpm : float, optional
            The minimization -- convergence break point for
            RTPM.
        fillna : int, optional
            Prevent division by zero in denominator causing nan
            in the optimization iterations.

        Attributes
        ----------
        eigvals : array-like
            The singular value vectors (1,n_components)
        proportion_explained : array-like
            The percent explained by each
            rank-1 factor. (1,n_components)
        loadings : list of array-like
            A list of loadings for all dimensions
            of the data
        subjects : array-like
            The sample loading vectors
            of shape (subjects, n_components)
        features : array-like
            The feature loading vectors
            of shape (features, n_components)
        conditions  : array-like or list of array-like
            The conditional loading vectors
            of shape (conditions, n_components) if there is 1 type
            of condition, and a list of such matrices if
            there are more than 1 type of condition
        subject_trajectory : list of array-like
            The loadings of the dot product between
            subjects and each condition in conditions.
            For each condition in conditions the
            loading is of shape (subjects, condition).
        feature_trajectory : list of array-like
            The loadings of the dot product between
            features and each condition in conditions.
            For each condition in conditions the
            loading is of shape (features, condition).
        subject_distances : list of array-like
            The euclidean distances of the subject trajectory
            for each condition in conditions.
            For each condition in conditions the
            loading is of shape (subjects, subjects).
        feature_distances : list of array-like
            The euclidean distances of the feature trajectory
            for each condition in conditions.
            For each condition in conditions the
            loading is of shape (features, features).
        dist : array-like
            A absolute distance vector
            between tensor and reconstructed tensor.

        References
        ----------
        .. [1] Jain, Prateek, and Sewoong Oh. 2014.
               “Provable Tensor Factorization with Missing Data.”
               In Advances in Neural Information Processing Systems
               27, edited by Z. Ghahramani, M. Welling, C. Cortes,
               N. D. Lawrence, and K. Q. Weinberger, 1431–39.
               Curran Associates, Inc.
        .. [2] A. Anandkumar, R. Ge, M., Janzamin,
               Guaranteed Non-Orthogonal Tensor
               Decomposition via Alternating Rank-1
               Updates. CoRR (2014),
               pp. 1-36.
        .. [3] A. Anandkumar, R. Ge, D. Hsu,
               S. M. Kakade, M. Telgarsky,
               Tensor Decompositions for Learning Latent Variable Models
               (A Survey for ALT). Lecture Notes in Computer Science
               (2015), pp. 19–38.

        Examples
        --------
        TODO
        """

        # save all input perameters as attributes
        self.n_components = n_components
        self.max_als_iterations = max_als_iterations
        self.max_rtpm_iterations = max_rtpm_iterations
        self.n_initializations = n_initializations
        self.tol_als = tol_als
        self.tol_rtpm = tol_rtpm
        self.fillna = fillna
        self.center = center
        self.check_dense = check_dense

    def fit(self, tensor, y=None):
        """

        Run _fit() a wrapper
        for the tenals helper.

        Parameters
        ----------
        tensor : array-like
            A tensor, often
            compositionally transformed,
            with missing values. The missing
            values must be zeros. Canonically,
            Tensor must be of the shape:
            first dimension = subjects
            second dimension = features
            rest dimensions = types of conditions
        """

        self.tensor = tensor.copy()
        self._fit()
        return self

    def _fit(self):
        """
        This function runs the
        tenals helper.

        """

        # make copy for imputation, check type
        tensor = self.tensor
        # ensure that the data is in the format of np.ndarray.
        if not isinstance(tensor, np.ndarray):
            raise ValueError('Input data is should be type numpy.ndarray')
        # ensure the data contains missing values.
        # other methods would be better in the case of fully dense data
        n_entries = np.product(tensor.shape)
        if self.check_dense:
            if (np.count_nonzero(tensor) == n_entries and
                    np.count_nonzero(~np.isnan(tensor)) == n_entries):
                err_ = 'No missing data in the format np.nan or 0.'
                raise ValueError(err_)
        # ensure there are no undefined values in the array
        if np.count_nonzero(np.isinf(tensor)) != 0:
            raise ValueError('Contains either np.inf or -np.inf')
        # ensure there are less components that the max. shape of the tensor
        if self.n_components > np.max(tensor.shape):
            raise ValueError(
                'n_components must be less than the maximum shape')
        # obtain mask array where values of tensor are zero
        mask = np.zeros(tensor.shape)  # mask of zeros the same shape
        mask[abs(tensor) > 0] = 1  # set masked (missing) values to one

        # run the tensor factorization method [2]
        loads, s, dist = tenals(tensor,
                                mask,
                                n_components=self.n_components,
                                n_initializations=self.n_initializations,
                                max_als_iterations=self.max_als_iterations,
                                max_rtpm_iterations=self.max_rtpm_iterations,
                                tol_als=self.tol_als,
                                tol_rtpm=self.tol_rtpm,
                                fillna=self.fillna)
        # save all raw laodings as attribute
        self.loadings = loads
        # the distance between tensor_imputed and tensor
        self.dist = dist
        # save array of loadings for subjects
        self.subjects = loads[0].copy()
        self.subjects = self.subjects[self.subjects[:, 0].argsort()]
        # save array of loadings for features
        self.features = loads[1].copy()
        self.features = self.features[self.features[:, 0].argsort()]
        # center the subject / feature biplot
        self.subjects -= self.subjects.mean(axis=0)
        self.features -= self.features.mean(axis=0)
        if self.center:
            # re-center using a final svd
            X = self.subjects @ s @ self.features.T
            possible_comp = [np.min(X.shape),
                             self.n_components]
            self.biplot_components = np.min(possible_comp)
            X = X - X.mean(axis=0)
            X = X - X.mean(axis=1).reshape(-1, 1)
            u, s, v = svd(X, full_matrices=False)
            u = u[:, :self.biplot_components]
            v = v.T[:, :self.biplot_components]
            p = s * (1 / s.sum())
            p = np.array(p[:self.biplot_components])
            s = np.diag(s[:self.biplot_components])
            # save the re-centered biplot
            self.features = v
            self.subjects = u
        else:
            # just make prop-exp
            p = np.array(np.diag(s) /
                         1 / np.sum(np.diag(s)))
        # save all eigen values
        self.eigvals = np.diag(s)
        # the proortion explained for n_components
        self.proportion_explained = p
        # save list of array(s) of loadings for conditions
        self.conditions = [loads[2]] if len(loads[2:]) == 1 \
            else loads[2:]
        # generate the trajectory(s) and distances
        # list of each condition-subject trajectory
        self.subject_trajectory = []
        # list of each condition-subject distance
        self.subject_distances = []
        # list of each condition-feature trajectory
        self.feature_trajectory = []
        # for each condition in conditions
        # generate a trajectory and distance array
        for condition in self.conditions:
            # temporary list of components in subject trajectory
            subject_temp_trajectory = []
            # temporary list of components in feature trajectory
            feature_temp_trajectory = []
            # for each component in the rank given to TensorFactorization
            for component in range(self.n_components)[::-1]:
                # component condition-subject trajectory
                dtmp = np.dot(loads[0][:, [component]],
                              condition[:, [component]].T).flatten()
                subject_temp_trajectory.append(dtmp)
                # component condition-feature trajectory
                dtmp = np.dot(loads[1][:, [component]],
                              condition[:, [component]].T).flatten()
                feature_temp_trajectory.append(dtmp)
            # combine all n_components
            subject_temp_trajectory = np.array(subject_temp_trajectory).T
            feature_temp_trajectory = np.array(feature_temp_trajectory).T
            # double check check centered
            subject_temp_trajectory -= subject_temp_trajectory.mean(axis=0)
            feature_temp_trajectory -= feature_temp_trajectory.mean(axis=0)
            # save subject-condition trajectory and distance matrix
            self.subject_trajectory.append(subject_temp_trajectory)
            self.subject_distances.append(
                distance.cdist(subject_temp_trajectory,
                               subject_temp_trajectory))
            # save feature-condition trajectory and distance matrix
            self.feature_trajectory.append(feature_temp_trajectory)

    def label(self,
              construct,
              taxonomy=None):
        """

        Label the loadings with the constructed tensor dimension
        labels from preprosessing.build.

        Parameters
        ----------
        construct : constructed object from preprosessing.build
            This object has the following attributes used here:
                construct.mapping : DataFrame
                    construct.mapping metadata used to build tensor
                    rows = samples
                    columns = categories
                construct.subject_order : list
                    order of subjects in tensor array
                construct.feature_order : list
                    order of features in tensor array
                construct.conditions : list of category names
                    category of conditional in metadata
                construct.condition_orders : list of lists
                    order of conditions for each
                    condition in tensor array
        taxonomy : DataFrame
            TODO
        """

        # columns labels PC1 ... PC(n_components)
        self.column_labels = ['PC' + str(i)
                              for i in range(1, self.n_components + 1)]
        self.biplot_labels = ['PC' + str(i)
                              for i in range(1, self.biplot_components + 1)]

        # % var explained
        self.proportion_explained = pd.Series(self.proportion_explained,
                                              index=self.biplot_labels)
        # eigvals
        self.eigvals = pd.Series(self.eigvals,
                                 index=self.biplot_labels)
        # DataFrame single non-condition dependent loadings
        self.subjects = pd.DataFrame(self.subjects,
                                     columns=self.biplot_labels,
                                     index=construct.subject_order)
        self.features = pd.DataFrame(self.features,
                                     columns=self.biplot_labels,
                                     index=construct.feature_order)

        # id taxonomy is given then add the taxonomic information
        if taxonomy is not None:
            self.features = pd.concat(
                [self.features, taxonomy], axis=1, sort=True)

        # list of DataFrame(s) for each condition loadings
        conditions = []
        for c_ind, condition in enumerate(construct.conditions):
            conditions.append(
                pd.DataFrame(
                    self.conditions[c_ind],
                    columns=self.column_labels,
                    index=construct.condition_orders[c_ind]))
        self.conditions = conditions

        # label the subject trajectories
        subject_trajectory = []
        for c_ind, condition in enumerate(construct.conditions):
            traj_tmp = self.subject_trajectory[c_ind]
            traj_tmp = pd.DataFrame(traj_tmp,
                                    columns=self.column_labels)
            ordr_ = [[subj, cond] for subj in construct.subject_order
                     for cond in construct.condition_orders[c_ind]]
            ordr_ = pd.DataFrame(ordr_, columns=['subject_id', condition])
            traj_tmp = pd.concat([traj_tmp, ordr_], axis=1)
            # add metadata for the trajectories
            index_map = construct.condition_metadata_map[c_ind]
            traj_tmp.index = [index_map[tuple(pair)]
                              if tuple(pair) in index_map.keys()
                              else '-'.join(list(map(str, pair)))
                              for pair in zip(traj_tmp['subject_id'].values,
                                              traj_tmp[condition].values)]
            traj_tmp.index.name = 'sample-id'
            subject_trajectory.append(traj_tmp)
        self.subject_trajectory = subject_trajectory

        # label the feature trajectories
        feature_trajectory = []
        for c_ind, condition in enumerate(construct.conditions):
            traj_tmp = self.feature_trajectory[c_ind]
            traj_tmp = pd.DataFrame(traj_tmp,
                                    columns=self.column_labels)
            ordr_ = [[feat, cond] for feat in construct.feature_order
                     for cond in construct.condition_orders[c_ind]]
            ordr_ = pd.DataFrame(ordr_, columns=['feature_id', condition])
            ordr_ = pd.concat([traj_tmp, ordr_], axis=1)
            if taxonomy is not None:
                if 'Taxon' in taxonomy.columns:
                    feat_map = dict(taxonomy.Taxon)
                    ordr_['Taxon'] = [feat_map[feat]
                                      if feat in feat_map.keys()
                                      else np.nan
                                      for feat in ordr_.feature_id]
                    # add taxonomic levels for grouping later (if available)

                    def tax_split(tax_id, tax_level): return tax_id.split(
                        tax_level)[1].split(';')[0]

                    for level, lname in zip(['k__', 'p__', 'c__', 'o__',
                                             'f__', 'g__', 's__'],
                                            ['kingdom', 'phylum', 'class',
                                             'order', 'family', 'genus',
                                             'species']):
                        if lname not in taxonomy.columns:
                            taxonomy_tmp = []
                            for tax in ordr_.Taxon:
                                if tax is not np.nan and\
                                   level in tax and\
                                   len(tax_split(tax, level)) > 0:
                                    taxonomy_tmp.append(tax_split(tax,
                                                                  level))
                                else:
                                    taxonomy_tmp.append(np.nan)
                            ordr_[lname] = taxonomy_tmp
                if 'Confidence' in taxonomy.columns:
                    feat_map = dict(taxonomy.Confidence)
                    ordr_['Confidence'] = [feat_map[feat]
                                           if feat in feat_map.keys()
                                           else np.nan
                                           for feat in ordr_.feature_id]

            ordr_.index.name = 'featureid'
            feature_trajectory.append(ordr_)
        self.feature_trajectory = feature_trajectory


def tenals(tensor,
           mask,
           n_components=3,
           n_initializations=50,
           max_als_iterations=50,
           max_rtpm_iterations=50,
           tol_als=1e-8,
           tol_rtpm=1e-8,
           fillna=1.0):
    """
    A low-rank 3rd order tensor factorization
    for partially observered non-symmetric
    sets. This method relies on a CANDECOMP/
    PARAFAC (CP) tensor decomposition. Missing
    values are handled by  and alternating
    least squares (ALS) minimization between
    tensor and reconstructed_tensor.

    Parameters
    ----------
    tensor : array-like
        A sparse `n` order tensor with zeros
        in place of missing values.
    mask : array-like
        A masking array of missing values.
    n_components : int, optional
        The underlying low-rank, will be
        equal to the number of rank 1
        components that are output. The
        higher the rank given, the more
        expensive the computation will
        be.
    max_als_iterations : int, optional
        Max number of Alternating Least Square (ALS).
    tol_als : float, optional
        The minimization -- convergence break point for
        ALS.
    max_rtpm_iterations : int, optional
        Max number of Robust Tensor Power Method (RTPM)
        iterations.
    n_initializations : int, optional
        The number of initial
        vectors. Larger values will
        give more accurate factorization
        but will be more computationally
        expensive.
    tol_rtpm : float, optional
        The minimization -- convergence break point for
        RTPM.
    fillna : int, optional
        Prevent division by zero in denominator causing nan
        in the optimization iterations.

    Returns
    -------
    loadings : list array-like
        The factors of tensor. The `i`th entry of loadings corresponds to
        the mode-`i` factors of tensor and hase shape (tensor.shape[i], r).
    eigvals : array-like
        The r-dimension vector of eigvals.
    dist : array-like
        A absolute distance vector
        between tensor and reconstructed tensor.

    Raises
    ------
    ValueError
        Nan values in input, factorization
        did not converge.

    References
    ----------
    .. [1] Jain, Prateek, and Sewoong Oh. 2014.
            “Provable Tensor Factorization with Missing Data.”
            In Advances in Neural Information Processing Systems
            27, edited by Z. Ghahramani, M. Welling, C. Cortes,
            N. D. Lawrence, and K. Q. Weinberger, 1431–39.
            Curran Associates, Inc.
    .. [2] A. Anandkumar, R. Ge, M., Janzamin,
            Guaranteed Non-Orthogonal Tensor
            Decomposition via Alternating Rank-1
            Updates. CoRR (2014),
            pp. 1-36.
    .. [3] A. Anandkumar, R. Ge, D. Hsu,
            S. M. Kakade, M. Telgarsky,
            Tensor Decompositions for Learning Latent Variable Models
            (A Survey for ALT). Lecture Notes in Computer Science
            (2015), pp. 19–38.

    Examples
    --------
    TODO
    """

    # Get the shape of tensor to iterate
    tensor_dimensions = tensor.shape
    # Frobenius norm initial for ALS minimization.
    initial_tensor_frobenius_norm = norm(tensor)**2
    # rank est. (for each slice)
    if len(tensor_dimensions) == 3:  # only for 3D
        for i in range(tensor_dimensions[0]):
            obs_tmp = tensor[i, :, :].copy()
            total_nonzeros = np.count_nonzero(mask[i, :, :].copy())
            n_, m_ = obs_tmp.shape
            eps_tmp = total_nonzeros / np.sqrt(n_ * m_)
            if min(obs_tmp.shape) <= 2:
                # two-subjects/time is already low-rank
                continue
            if rank_estimate(obs_tmp, eps_tmp) >= (min(obs_tmp.shape) - 1):
                warnings.warn('A component of your data may be high-rank.',
                              RuntimeWarning)
    # initial by Robust Tensor Power Method (modified for non-symmetric
    # tensors).
    initial_eigvals, initial_loadings = robust_tensor_power_method(
        tensor, n_components, n_initializations, max_rtpm_iterations, tol_rtpm)
    # Begin alternating least squares minimization below!
    # make a copy of inital factorization from RTPM
    # to use in the ALS step.
    loadings = initial_loadings.copy()
    eigvals = initial_eigvals.copy()
    # Iterate minimization for at maximum max_als_iterations
    # can break if converges and satisfies tol_als.
    for iteration in range(max_als_iterations):
        # For each rank-1 vector in n_components (total rank)
        for component in range(n_components):
            # set all eigvals to zero for iteration
            eigvals[component] = 0
            # reconstruct the tensor from the loadings to compare to the
            # original tensor
            reconstructed_tensor = construct_tensor(eigvals, loadings)
            reconstructed_tensor = np.multiply(reconstructed_tensor, mask)
            # generate copy of loadings to reconstruct on each iterations
            loadings_iter = [loading[:, component].copy()
                             for loading in loadings]
            # denominator should end up as a list of np.zeros((dim_i, 1))
            denominator = [np.zeros(dim) for dim in tensor_dimensions]
            # for each  dimension of the tensor optimize that dimensions
            # loading based on the distance between the original tensor
            # and the reconstruction
            for dim, dim_size in enumerate(tensor_dimensions):
                # set previous loading to zero
                loadings[dim][:, component] = 0
                # generate indices to perform dot-product across
                dims_np = np.arange(len(tensor_dimensions))
                dot_across = dims_np[dims_np != dim]
                # outer dot-product
                err = tensor - reconstructed_tensor
                l_iter = loadings_iter[dot_across[0]]
                axes = (1 if dim == 0 else 0, 0)
                construct_loadings = np.tensordot(err,
                                                  l_iter,
                                                  axes=axes)
                denominator[dim] = np.tensordot(mask,
                                                l_iter**2,
                                                axes=axes)
                # inner dot-product
                for inner_dim in dot_across[1:]:
                    l_iter = loadings_iter[inner_dim]
                    axes = (1 if inner_dim > dim else 0, 0)
                    construct_loadings = np.tensordot(construct_loadings,
                                                      l_iter, axes=axes)
                    denominator[dim] = np.tensordot(denominator[dim],
                                                    l_iter**2, axes=axes)
                # update iteration's loadings
                loadings_iter[dim] = loadings[dim][:, component] +\
                    construct_loadings.flatten()
                # Add fillna to prevent division by zero in denominator
                # causing nan. This can occur in early iteration from all
                # zero fibers in the tensor along dim. In practice this
                # should be rare but can occur in very sparse tensors.
                denominator[dim][denominator[dim] == 0] = fillna
                loadings_iter[dim] = loadings_iter[dim] / \
                    denominator[dim]
                # If  this dimension is the last in the tensor then
                # update the eigvals with the new loadings.
                if dim == len(tensor_dimensions) - 1:
                    eigvals[component] = norm(loadings_iter[dim])
                #  update the loadings_iter in dimension (dim)
                eigvals[component][eigvals[component] == 0] = fillna
                loadings_iter[dim] = loadings_iter[dim] / \
                    eigvals[component]
                loadings[dim][:, component] = loadings_iter[dim]
            # update the loadings with this iterations loadings
            for i, loading in enumerate(loadings):
                loading[:, component] = loadings_iter[i]
        # MSE of the original tensor and the reconstructed tensor
        # based on the loadings.
        mean_squared_error = tensor - mask * \
            construct_tensor(eigvals, loadings)
        # Frobenius norm for new reconstructed tensor
        iteration_tensor_frobenius_norm = norm(mean_squared_error)**2
        # If the error between this iterations reconstruction and the
        # intital tensor is below tol_als then  break the iterations.
        err_conv = np.sqrt(iteration_tensor_frobenius_norm /
                           initial_tensor_frobenius_norm)
        if err_conv < tol_als:
            break
    # check that the factorization converged
    if any(sum(sum(np.isnan(loading))) > 0 for loading in loadings):
        raise ValueError("The factorization did not converge.",
                         "Please check the input tensor for errors.")
    # Get the diagonal of the eigvals
    eigvals = np.diag(eigvals.flatten())
    # Sort the eigvals and the laodings from
    # largest to smallest.
    idx = np.argsort(np.diag(eigvals))[::-1]
    eigvals = eigvals[idx, :][:, idx]
    loadings = [loading[:, idx[::-1]]
                for loading in loadings]

    return loadings, eigvals, err_conv


def robust_tensor_power_method(tensor,
                               n_components,
                               n_initializations=50,
                               max_rtpm_iterations=50,
                               tol_rtpm=1e-10,
                               random_state=42):
    """
    The Robust Tensor Power Method
    (RTPM). Is a generalization of
    the widely used power method for
    computing lead singular values
    of a matrix and can approximate
    the largest singular vectors of
    a tensor.

    Parameters
    ----------
    tensor : array-like
        A sparse `n` order tensor with zeros
        in place of missing values.
    n_components : int, optional
        The underlying low-rank, will be
        equal to the number of rank 1
        components that are output. The
        higher the rank given, the more
        expensive the computation will
        be.
    n_initializations : int, optional
        The number of initial
        vectors. Larger values will
        give more accurate factorization
        but will be more computationally
        expensive.
    max_rtpm_iterations : int, optional
        Max number of iterations.
    tol_rtpm : flat, optional
        The minimization -- convergence break point for
        RTPM.

    Returns
    -------
    eigvals : array-like
        The eigvals of the factorizations
    loadings : list of array-like
        The `i`-th entry of initial_loadings corresponds to
        the factors along the `i`-th mode of tensor

    References
    ----------
    .. [2] A. Anandkumar, R. Ge, M., Janzamin,
            Guaranteed Non-Orthogonal Tensor
            Decomposition via Alternating Rank-1
            Updates. CoRR (2014),
            pp. 1-36.

    """

    # tensor shape is the number of loadings
    dims = tensor.shape
    # for each dim. initalize a loading fiber
    loadings = [np.ones((n, n_components)) for n in dims]
    # initlize n_component eigvals
    eigvals = np.ones((n_components, 1))
    # iterate on loadings from first to last
    for r in range(n_components):
        # loadings and eigs to fille for
        # each random initalization vector
        tU = [np.zeros((n, n_initializations)) for n in dims]
        tS = np.zeros((n_initializations, 1))
        # random initialization
        if random_state is None or isinstance(random_state, int):
            rnd = np.random.RandomState(random_state)
        elif isinstance(random_state, np.random.RandomState):
            rnd = random_state
        else:
            raise ValueError('Random State must be of type ',
                             'np.random.RandomState or int.')
        # run n-initializations
        for init in range(n_initializations):
            # single random initialization
            init_load = [rnd.random_sample((n, 1)) for n in dims]
            init_load = [vec / norm(vec) for vec in init_load]
            # construct tensor
            T_con = construct_tensor(eigvals, init_load)
            # run power iteration's for loading initialization
            initializations = asymmetric_power_update(tensor - T_con,
                                                      init_load,
                                                      max_rtpm_iterations,
                                                      tol_rtpm)
            # fill that initalization (looking for maximum)
            for idx, load_tmp in enumerate(initializations):
                tU[idx][:, init] = load_tmp
                tU[idx][:, init] = tU[idx][:, init] / norm(tU[idx][:, init])
            # generate eigvals for that initalization
            T_con = construct_tensor(eigvals, loadings)
            tS[init] = eigval_update(tensor - T_con,
                                     [ltmp[:, [init]] for ltmp in tU])
        # find the maximum (absolute) eigval
        max_idx = np.argmax(abs(tS))
        # for that optimal eigval add it to
        # the final laodings & eigvals.
        for idx, max_load in enumerate(tU):
            loadings[idx][:, r] = max_load[:, max_idx]
            loadings[idx][:, r] = (loadings[idx][:, r]
                                   / norm(loadings[idx][:, r]))
        # fill eigval
        T_con = construct_tensor(eigvals, loadings)
        eigvals[r] = eigval_update(tensor - T_con,
                                   [ltmp[:, [r]] for ltmp in loadings])

    return eigvals, loadings


def asymmetric_power_update(tensor,
                            init,
                            max_rtpm_iterations=50,
                            tol_rtpm=1e-10):
    """
    Completes a single iteration of optimization
    for a random start of RTPM

    Parameters
    ----------
    tensor : array-like
        an `n` order tensor
    init : list of array-like
        randomly generated loadings
        for each dimension of tensor
    max_rtpm_iterations : int
        maximum iterations.

    Returns
    -------
    list of array-like
        entry `i` of list is a single
        vector corresponding to the `i`th
        mode of `tensor`

    """

    # tensor dims is number of loadings
    n_dims = len(tensor.shape)
    # begin power iterations
    for itr in range(max_rtpm_iterations):
        loadings = []
        dims = np.arange(n_dims)
        # iterate on each loading (dim.)
        for dim in dims:
            # generate updates for each loading dimention
            dot_across = dims[dims != dim]
            axis = (1 if dim == 0 else 0, 0)
            load_dim = np.tensordot(tensor,
                                    init[dot_across[0]],
                                    axes=axis)
            for inner_dim in dot_across[1:]:
                axis = (1 if inner_dim > dim else 0, 0)
                load_dim = np.tensordot(load_dim,
                                        init[inner_dim],
                                        axes=axis)
            loadings.append(load_dim)
        # fill the update loadings
        new_shapes = [u_n.shape[:(-1 * (len(dims) - 2))]
                      for u_n in loadings]
        loadings = [u_n.reshape(new_shape)
                    for u_n, new_shape in zip(loadings,
                                              new_shapes)]
        # update initalized loading and keep prev.
        init_prev = [u for u in init]
        init = [u_i / norm(u_i) for u_i in loadings]
        # calulate the iteration minimization (exit case for tol_rtpm)
        tol_itr = sum(norm(u0 - u) for u0, u in zip(init_prev, init))
        if tol_itr < tol_rtpm:
            break
    return [loading.flatten() for loading in init]


def construct_tensor(S, U):
    """
    This function takes the
    CP decomposition of a 3rd
    order tensor and outputs
    the reconstructed tensor
    TE_hat.
    Parameters
    ----------
    S : array-like
        The r-dimension vector.
    U : list of array-like
        Element i is a factor of shape
        (n[i], r).
    Returns
    -------
    T : array-like
        TE_hat of shape
        tuple(n[i] for i in range(len(U))).
    """

    output_shape = tuple(u.shape[0] for u in U)
    to_multiply = [S.T * u if i == 0 else u for i, u in enumerate(U)]
    product = khatri_rao(to_multiply)
    T = product.sum(1).reshape(output_shape)

    return T


def eigval_update(D, U_list):
    """
    The Orthogonal tensor
    projection created by
    the TE - TE_hat distance.
    Used in the initialization
    step with RTPM_single.
    Parameters
    ----------
    D : array-like
        with shape (n[0], n[1], ..., )
    U_list : list of array-like
        Element i is a factor of shape
        (n[i], r). Same length as D.shape
    Returns
    -------
    M : float
        The multilinear mapping of D on U_list
    """
    current = D
    for u in U_list:
        current = np.tensordot(current, u, axes=(0, 0))
    return current


def khatri_rao(matrices):
    """
    Returns the Khatri Rao product of a list of matrices
    Modified from TensorLy
    Parameters
    ----------
    matrices : list of array-like
        Matrices to take the Khatri Rao Product of
    Returns
    -------
    array-like
        The Khatri Rao Product of the matrices in `matrices`
    References
    ----------
    .. [1] Jean Kossaifi, Yannis Panagakis, Anima Anandkumar and Maja
            Pantic, TensorLy: Tensor Learning in Python,
            https://arxiv.org/abs/1610.09555.
    """

    n_columns = matrices[0].shape[1]
    n_factors = len(matrices)

    start = ord('a')
    common_dim = 'z'
    target = ''.join(chr(start + i) for i in range(n_factors))
    source = ','.join(i + common_dim for i in target)
    operation = source + '->' + target + common_dim

    return np.einsum(operation, *matrices).reshape((-1, n_columns))



import unittest
import numpy as np
import pandas as pd
import t2t.nlevel as nl
from biom import Table
from skbio import TreeNode
import numpy.testing as npt
from skbio.stats.composition import clr
from skbio.util import get_data_path
from pandas.testing import assert_frame_equal
from gemelli.preprocessing import (build, tensor_rclr,
                                   matrix_closure,
                                   matrix_rclr,
                                   bp_read_phylogeny,
                                   fast_unifrac,
                                   tree_topology_filter,
                                   TaxonomyError,
                                   create_taxonomy_metadata,
                                   retrieve_t2t_taxonomy,
                                   _pull_consensus_strings,
                                   _get_taxonomy_io_stream)


class Testpreprocessing(unittest.TestCase):

    def setUp(self):
        # test closeure
        self.close_table_zero = np.array([[2, 2, 6],
                                          [0, 0, 0]])
        self.close_true_zero = np.array([[0.2,  0.2,  0.6],
                                         [np.nan] * 3])
        self.close_table = np.array([[2, 2, 6],
                                     [2, 2, 6]])
        self.close_true = np.array([[0.2,  0.2,  0.6],
                                    [0.2,  0.2,  0.6]])
        # matrix_rclr base tests
        self.cdata1 = np.array([[2, 2, 6],
                                [4, 4, 2]])
        self.cdata2 = np.array([[3, 3, 0],
                                [0, 4, 2]])
        self.true2 = np.array([[0.0, 0.0, np.nan],
                               [np.nan, 0.34657359, -0.34657359]])
        self.bad1 = np.array([1, 2, -1])
        self.bad2 = np.array([1, 2, np.inf])
        self.bad3 = np.array([1, 2, np.nan])
        # test dense
        self.count_data_one = np.array([[2, 2, 6],
                                        [4, 4, 2]])
        # test with zeros
        self.count_data_two = np.array([[3, 3, 0],
                                        [0, 4, 2]])
        # test dense tensor
        self.tensor_true = np.array([[[4., 5., 6.],
                                      [1., 2., 3.]],
                                     [[10., 11., 12.],
                                      [7., 8., 9.]],
                                     [[16., 17., 18.],
                                      [13., 14., 15.]]])
        # make a test tree
        newick = ('((((A:1,B:1)n9:3,C:1)n8:3,(D:1,E:1)n7:2)n4:3'
                  ',((F:1,G:1)n6:4,(H:1)n5:4)n3:3)n1;')
        self.tree = TreeNode.read([newick])
        # make a test tree to raise error
        newick_bad = ('((((A,B)n9,C)n8,(D,E)n7)n4'
                      ',((F,G)n6,(H)n5)n3)n1;')
        self.tree_bad = TreeNode.read([newick_bad])
        # make a test for reading off consensus
        tax_tree = "(((a,b,c)n1)n2,(d,e)n3)n4;"
        self.tax_tree = TreeNode.read([tax_tree])
        consensus_tree = "(((a,b,c)'s__3')'p__1; s__3',(d,e)'p__2')'k__0';"
        self.consensus_tree = TreeNode.read([consensus_tree])
        self.consensus_string = ['k__0; p__; s__',
                                 'k__0; p__1; s__3',
                                 'k__0; p__1; s__3',
                                 'k__0; p__1; s__3',
                                 'k__0; p__1; s__3',
                                 'k__0; p__1; s__3',
                                 'k__0; p__2; s__',
                                 'k__0; p__2; s__',
                                 'k__0; p__2; s__']
        # make a test taxonomy
        taxonomy = {'Feature ID': ['a', 'b', 'c', 'd', 'e'],
                    'Taxon': ['k__0; p__1; s__',
                              'k__0; p__1; s__3',
                              'k__0; p__1; s__3',
                              'k__0; p__2; s__',
                              'k__0; p__2; s__']}
        self.taxonomy = pd.DataFrame(data=taxonomy).set_index('Feature ID')
        self.taxonomy_io_stream_contents = ['a\tk__0; p__1; s__\n',
                                            'b\tk__0; p__1; s__3\n',
                                            'c\tk__0; p__1; s__3\n',
                                            'd\tk__0; p__2; s__\n',
                                            'e\tk__0; p__2; s__\n']
        t2t_taxonomy = {
            'Feature ID': ['n4', 'n2', 'n1', 'a', 'b', 'c', 'n3', 'd', 'e'],
            'Taxon': ['k__0; p__; s__',
                      'k__0; p__1; s__3',
                      'k__0; p__1; s__3',
                      'k__0; p__1; s__3',
                      'k__0; p__1; s__3',
                      'k__0; p__1; s__3',
                      'k__0; p__2; s__',
                      'k__0; p__2; s__',
                      'k__0; p__2; s__']

        }
        self.t2t_taxonomy = pd.DataFrame(data=t2t_taxonomy)\
            .set_index('Feature ID')
        # make a test taxonomy to raise error
        bad_taxonomy = {'Feature ID': ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H'],
                        'bad_column': ['k__k; p__a; s__a',
                                       'k__k; p__a; s__a',
                                       'k__k; p__a; s__a',
                                       'k__k; p__a; s__b',
                                       'k__k; p__a; s__b',
                                       'k__k; p__b; s__c',
                                       'k__k; p__b; s__c',
                                       'k__k; p__b; s__d']}
        self.bad_taxonomy = pd.DataFrame(data=bad_taxonomy)
        # make a test table
        counts = np.array([[2, 0, 3, 3, 1, 2, 3, 4],
                           [2, 1, 4, 1, 0, 1, 4, 0]]).T
        feature_ids = [tip_.name for tip_ in self.tree.tips()]
        subject_ids = ['s1', 's2']
        self.phylo_table = Table(counts, feature_ids, subject_ids)
        # fully labeled tree
        self.ids_true = ['A', 'B', 'n2', 'C',
                         'D', 'E', 'n6', 'n7',
                         'F', 'G', 'H', 'n11', 'n12',
                         'n13', 'n14', 'n15']
        # true lengths
        self.branch_length_true = np.array([1., 1., 3., 1., 1., 1., 3., 2.,
                                           1., 1., 1., 4., 4., 3., 3., 0.])
        # true vector count table
        self.vector_counts_true = np.array([[2,  0,  2,  3,  3,  1,  5,  4,
                                            2,  3,  4,  5,  4,  9,  9, 18],
                                            [2,  1,  3,  4,  1,  0,  7,  1,
                                            1, 4,  0,  5,  0,  8,  5, 13]])
        pass

    def test_closure_missing(self):
        """Test closure with zeros (due to tensor)."""
        # test a case with zeros
        cmat_res = matrix_closure(self.close_table_zero)
        npt.assert_allclose(cmat_res, self.close_true_zero)

    def test_closure(self):
        """Test closure without zeros."""
        # test a case with zeros
        cmat_res = matrix_closure(self.close_table)
        npt.assert_allclose(cmat_res, self.close_true)

    def test_rclr_sparse(self):
        """Test matrix_rclr on sparse data."""
        # test a case with zeros
        cmat = matrix_rclr(self.cdata2)
        npt.assert_allclose(cmat, self.true2)

    def test_rclr_negative_raises(self):
        """Test matrix_rclr ValueError on negative."""
        # test negatives throw value error
        with self.assertRaises(ValueError):
            matrix_rclr(self.bad1)

    def test_rclr_inf_raises(self):
        """Test matrix_rclr ValueError on undefined."""
        # test undefined throw value error
        with self.assertRaises(ValueError):
            matrix_rclr(self.bad2)

    def test_rclr_nan_raises(self):
        """Test matrix_rclr ValueError on missing (as nan)."""
        # test nan throw value error
        with self.assertRaises(ValueError):
            matrix_rclr(self.bad3)

    def test_fast_unifrac_branch_raises(self):
        """Test fast_unifrac ValueError on tree with no branch lengths."""
        # test nan throw value error
        with self.assertRaises(ValueError):
            _ = fast_unifrac(self.phylo_table, self.tree_bad)

    def test_bp_read_phylogeny_mindepth_raises(self):
        """Test bp_read_phylogeny ValueError min_depth too large."""
        # test nan throw value error
        with self.assertRaises(ValueError):
            _ = bp_read_phylogeny(self.phylo_table,
                                  get_data_path('test_tree.nwk',
                                                subfolder='data'),
                                  min_depth=8)

    def test_fast_unifrac(self):
        """Test fast_unifrac table vectorized on tree."""
        # run vectorized table
        tmp_res = fast_unifrac(self.phylo_table, self.tree)
        counts_res, _, branch_lengths_res, fids_res, _ = tmp_res
        # test all expected
        npt.assert_allclose(counts_res,
                            self.vector_counts_true)
        npt.assert_allclose(branch_lengths_res,
                            self.branch_length_true)
        self.assertListEqual(fids_res, self.ids_true)

    def test_tree_topology_filter(self):
        """
        Test tree metric calculations.
        Original function comes from
        https://github.com/biocore/wol
        kindly provided here by Qiyun Zhu.
        Example from Fig. 9a of Puigbo, et al., 2009, J Biol.
                                                /-A
                                      /n9------|
                            /n8------|          \\-B
                           |         |
                  /n4------|          \\-C
                 |         |
                 |         |          /-D
                 |          \n7------|
                 |                    \\-E
                 |
                 |                    /-F
        -n1------|          /n6------|
                 |         |          \\-G
                 |-n3------|
                 |         |          /-H
                 |          \n5------|
                 |                    \\-I
                 |
                 |          /-J
                  \n2------|
                            \\-K
        """
        tree = TreeNode.read([
            '((((A,B)n9,C)n8,(D,E)n7)n4,((F,G)n6,(H,I)n5)n3,(J,K)n2)n1;'
        ])
        tree_topology_filter(tree)
        obs = {x.name: [getattr(x, y) for y in
                        ('n')]
               for x in tree.traverse()}
        exp = {
            'n1': [11],
            'n4': [5],
            'n3': [4],
            'n2': [2],
            'n8': [3],
            'n7': [2],
            'n6': [2],
            'n5': [2],
            'J': [1],
            'K': [1],
            'n9': [2],
            'C': [1],
            'D': [1],
            'E': [1],
            'F': [1],
            'G': [1],
            'H': [1],
            'I': [1],
            'A': [1],
            'B': [1]
        }
        self.assertDictEqual(obs, exp)

    def test_build(self):
        """Test building a tensor from metadata (multi-mode) & matrix_rclr."""
        # flatten tensor into matrix
        matrix_counts = self.tensor_true.transpose([0, 2, 1])
        reshape_shape = matrix_counts.shape
        matrix_counts = matrix_counts.reshape(9, 2)
        # build mapping and table dataframe to rebuild
        mapping = np.array([[0, 0, 0, 1, 1, 1, 2, 2, 2],
                            [0, 1, 2, 0, 1, 2, 0, 1, 2]])
        mapping = pd.DataFrame(mapping.T,
                               columns=['ID', 'conditional'])
        table = pd.DataFrame(matrix_counts.T)
        # rebuild the tensor
        tensor = build()
        tensor.construct(table, mapping,
                         'ID', ['conditional'])
        # ensure rebuild tensor is the same as it started
        npt.assert_allclose(tensor.counts,
                            self.tensor_true.astype(float))
        # test tensor is ordered correctly in every dimension
        self.assertListEqual(tensor.subject_order,
                             list(range(3)))
        self.assertListEqual(tensor.feature_order,
                             list(range(2)))
        self.assertListEqual(tensor.condition_orders[0],
                             list(range(3)))
        # test that flattened matrix has the same clr
        # transform as the tensor tensor_rclr
        tensor_clr_true = clr(matrix_counts).reshape(reshape_shape)
        tensor_clr_true = tensor_clr_true.transpose([0, 2, 1])
        npt.assert_allclose(tensor_rclr(tensor.counts),
                            tensor_clr_true)

    def test_errors(self):
        """Test building a tensor error raises."""
        # flatten tensor into matrix
        matrix_counts = self.tensor_true.transpose([0, 2, 1])
        matrix_counts = matrix_counts.reshape(9, 2)
        # build mapping and table dataframe to rebuild
        mapping = np.array([[0, 0, 0, 1, 1, 1, 2, 2, 2],
                            [0, 1, 2, 0, 1, 2, 0, 1, 2]])
        mapping = pd.DataFrame(mapping.T,
                               columns=['ID', 'conditional'])
        table = pd.DataFrame(matrix_counts.T)
        # rebuild the tensor
        tensor = build()
        tensor.construct(table, mapping,
                         'ID', ['conditional'])
        # test less than 2D throws ValueError
        with self.assertRaises(ValueError):
            tensor_rclr(np.array(range(3)))
        # test negatives throws ValueError
        with self.assertRaises(ValueError):
            tensor_rclr(tensor.counts * -1)
        tensor_true_error = self.tensor_true.astype(float)
        tensor_true_error[tensor_true_error <= 10] = np.inf
        # test infs throws ValueError
        with self.assertRaises(ValueError):
            tensor_rclr(tensor_true_error)
        tensor_true_error = self.tensor_true.astype(float)
        tensor_true_error[tensor_true_error <= 10] = np.nan
        # test nan(s) throws ValueError
        with self.assertRaises(ValueError):
            tensor_rclr(tensor_true_error)
        # test matrix_rclr on already made tensor
        with self.assertRaises(ValueError):
            matrix_rclr(self.tensor_true)
        # test matrix_rclr on negatives
        with self.assertRaises(ValueError):
            matrix_rclr(self.tensor_true * -1)
        # test that missing id in mapping ValueError
        with self.assertRaises(ValueError):
            tensor.construct(table, mapping.drop(['ID'], axis=1),
                             'ID', ['conditional'])
        # test that missing conditional in mapping ValueError
        with self.assertRaises(ValueError):
            tensor.construct(table, mapping.drop(['conditional'], axis=1),
                             'ID', ['conditional'])
        # test negatives throws ValueError
        with self.assertRaises(ValueError):
            tensor.construct(table * -1, mapping,
                             'ID', ['conditional'])
        table_error = table.astype(float)
        table_error[table_error <= 10] = np.inf
        # test infs throws ValueError
        with self.assertRaises(ValueError):
            tensor.construct(table_error, mapping,
                             'ID', ['conditional'])
        table_error = table.astype(float)
        table_error[table_error <= 10] = np.nan
        # test nan(s) throws ValueError
        with self.assertRaises(ValueError):
            tensor.construct(table_error, mapping,
                             'ID', ['conditional'])
        # test adding up counts for repeat samples
        table[9] = table[8] - 1
        mapping.loc[9, ['ID', 'conditional']
                    ] = mapping.loc[8, ['ID', 'conditional']]
        with self.assertWarns(Warning):
            tensor.construct(table, mapping, 'ID', ['conditional'])
        duplicate_tensor_true = self.tensor_true.copy()
        duplicate_tensor_true[2, :, 2] = duplicate_tensor_true[2, :, 2] - 1
        npt.assert_allclose(tensor.counts,
                            duplicate_tensor_true.astype(float))

    def test_matrix_tensor_rclr(self):
        """Test matrix == tensor matrix_rclr."""
        # test clr works the same if there are no zeros
        npt.assert_allclose(tensor_rclr(self.count_data_one.T).T,
                            clr(self.count_data_one))
        # test a case with zeros
        tensor_rclr(self.count_data_two)
        # test negatives throw ValueError
        with self.assertRaises(ValueError):
            tensor_rclr(self.tensor_true * -1)

    def test_rclr_dense(self):
        """Test matrix_rclr and clr are the same on dense datasets."""
        # test clr works the same if there are no zeros
        cmat = matrix_rclr(self.cdata1)
        npt.assert_allclose(cmat, clr(self.cdata1.copy()))

    def test_get_taxonomy_io_stream_invalid_column_raise(self):
        """Test _get_taxonomy_io_stream TaxonomyError invalid index."""
        with self.assertRaises(TaxonomyError):
            _ = _get_taxonomy_io_stream(self.bad_taxonomy)

    def test_get_taxonomy_io_stream(self):
        io_stream = _get_taxonomy_io_stream(self.taxonomy)
        stream_contents = io_stream.readlines()
        self.assertListEqual(stream_contents, self.taxonomy_io_stream_contents)

    def test_pull_consensus_strings(self):
        nl.set_rank_order(['k', 'p', 's'])
        self.assertListEqual(_pull_consensus_strings(self.consensus_tree),
                             self.consensus_string)

    def test_retrieve_t2t_taxonomy(self):
        self.assertListEqual(
            retrieve_t2t_taxonomy(self.tax_tree, self.taxonomy),
            self.consensus_string)

    def test_create_taxonomy_metadata(self):
        trav_tax = retrieve_t2t_taxonomy(self.tax_tree, self.taxonomy)
        assert_frame_equal(
            create_taxonomy_metadata(self.tax_tree, trav_tax),
            self.t2t_taxonomy)


if __name__ == "__main__":
    unittest.main()



import unittest
from os.path import sep as os_path_sep
import numpy as np
import pandas as pd
from pandas import read_csv
from biom import load_table, Table
from skbio import OrdinationResults, TreeNode
from skbio.stats.distance import DistanceMatrix
from skbio.util import get_data_path
from qiime2 import Artifact
from qiime2 import Metadata
from qiime2.plugins import gemelli as q2gemelli
from gemelli.ctf import ctf
from gemelli.scripts._standalone_ctf import (standalone_ctf,
                                             standalone_phylogenetic_ctf)
from numpy.testing import (assert_allclose, assert_array_almost_equal)
from gemelli.testing import absolute_sort
from click.testing import CliRunner
from nose.tools import nottest


@nottest
def create_test_table(include_phylogeny=False):

    if include_phylogeny:
        in_table = get_data_path('test-small.biom')
        in_meta = get_data_path('test-small.tsv')
        in_tree = get_data_path('test-small-tree.nwk')
        return in_table, in_meta, in_tree
    else:
        in_table = get_data_path('test-small.biom')
        in_meta = get_data_path('test-small.tsv')
        return in_table, in_meta


class Testctf(unittest.TestCase):

    def setUp(self):
        self.in_table, self.in_meta = create_test_table()
        self.subj = 'host_subject_id'
        self.state = 'context'
        self.biom_table = load_table(self.in_table)
        self.meta_table = read_csv(self.in_meta, sep='\t', index_col=0)
        # make metadata subjects and two-time
        self.meta_table_two_time = self.meta_table.iloc[:22, :].copy()
        self.meta_table_two_time[self.state] = [0] * 11 + [1] * 11
        self.meta_table_two_time[self.subj] = [i for i in range(11)] * 2

    def test_ctf(self):
        """Tests the basic validity of the actual ctf() method's outputs."""
        for meta_classes in [self.meta_table, self.meta_table_two_time]:
            res_tmp = ctf(table=self.biom_table,
                          sample_metadata=meta_classes,
                          individual_id_column=self.subj,
                          state_column=self.state)
            ord1, ord2, disttst, stst, ftst = res_tmp
            # Validate types of the ctf outputs
            self.assertIsInstance(ord1, OrdinationResults)
            self.assertIsInstance(ord2, OrdinationResults)
            self.assertIsInstance(disttst, DistanceMatrix)
            self.assertIsInstance(stst, pd.DataFrame)
            self.assertIsInstance(ftst, pd.DataFrame)
            # Ensure that no NaNs are in the OrdinationResults
            # NOTE that we have to use the DataFrame
            # .any() functions instead of
            # python's built-in any() functions --
            # see #29 for details on this
            self.assertFalse(np.isnan(ord1.features).any(axis=None))
            self.assertFalse(np.isnan(ord1.samples).any(axis=None))
            self.assertFalse(np.isnan(ord2.features).any(axis=None))
            self.assertFalse(np.isnan(ord2.samples).any(axis=None))


class Test_qiime2_ctf(unittest.TestCase):

    def setUp(self):
        self.subj = 'host_subject_id'
        self.state = 'context'
        paths_ = create_test_table(include_phylogeny=True)
        self.in_table, self.in_meta, self.in_tree = paths_
        self.biom_table = load_table(self.in_table)
        self.phylogeny = TreeNode.read(self.in_tree,
                                       format='newick')
        self.q2phylogeny = Artifact.import_data("Phylogeny[Rooted]",
                                                self.phylogeny)
        self.q2table = Artifact.import_data("FeatureTable[Frequency]",
                                            self.biom_table)
        self.meta_table = read_csv(self.in_meta, sep='\t', index_col=0)
        # make metadata subjects and two-time
        self.meta_table_two_time = self.meta_table.iloc[:22, :].copy()
        self.meta_table_two_time[self.state] = [0] * 11 + [1] * 11
        self.meta_table_two_time[self.subj] = [i for i in range(11)] * 2
        self.q2meta = Metadata(self.meta_table)
        self.q2meta_two_time = Metadata(self.meta_table)
        self.out_ = os_path_sep.join(self.in_table.split(os_path_sep)[:-1])

    def test_qiime2_ctf(self):
        """Tests that the Q2 and standalone ctf results match.

           Also validates against ground truth "expected" results.
        """

        # Run gemelli through QIIME 2 (specifically, the Artifact API)
        res = q2gemelli.actions.ctf(table=self.q2table,
                                    sample_metadata=self.q2meta,
                                    individual_id_column=self.subj,
                                    state_column=self.state)
        oqza, osqza, dqza, sqza, fqza = res
        # Get the underlying data from these artifacts
        q2straj = sqza.view(pd.DataFrame)
        q2ftraj = fqza.view(pd.DataFrame)

        # Next, run gemelli outside of QIIME 2. We're gonna check that
        # everything matches up.
        # ...First, though, we need to write the contents of self.q2table to a
        # BIOM file, so gemelli can understand it.
        # Derived from a line in test_standalone_ctf()
        # Run gemelli outside of QIIME 2...
        runner = CliRunner()
        result = runner.invoke(standalone_ctf,
                               ['--in-biom',
                                self.in_table,
                                '--sample-metadata-file',
                                self.in_meta,
                                '--individual-id-column',
                                'host_subject_id',
                                '--state-column-1',
                                'context',
                                '--output-dir',
                                self.out_])
        # check that exit code was 0 (indicating success)
        try:
            self.assertEqual(0, result.exit_code)
        except AssertionError:
            ex = result.exception
            error = Exception('Command failed with non-zero exit code')
            raise error.with_traceback(ex.__traceback__)
        # ...and read in the resulting output files. This code was derived from
        # test_standalone_ctf() elsewhere in gemelli's codebase.
        samp_res = read_csv(
            get_data_path('context-subject-ordination.tsv'),
            sep='\t',
            index_col=0)
        feat_res = read_csv(
            get_data_path('context-features-ordination.tsv'),
            sep='\t',
            index_col=0)
        # Check that the trajectory matrix matches our expectations
        comp_col = ['PC1', 'PC2', 'PC3']
        assert_allclose(absolute_sort(samp_res[comp_col].values),
                        absolute_sort(q2straj[comp_col].values),
                        atol=.5)
        assert_allclose(absolute_sort(feat_res[comp_col].values),
                        absolute_sort(q2ftraj[comp_col].values),
                        atol=.5)

    def test_qiime2_phylogenetic_ctf(self):
        """Tests that the Q2 & standalone phylogenetic ctf results match.

           Also validates against ground truth "expected" results.
        """

        # Run gemelli through QIIME 2 (specifically, the Artifact API)
        res = q2gemelli.actions.phylogenetic_ctf_without_taxonomy(self.q2table,
                                                 self.q2phylogeny,
                                                 self.q2meta,
                                                 self.subj,
                                                 self.state)
        oqza, osqza, dqza, sqza, fqza, tree, ctable, _ = res
        # Get the underlying data from these artifacts
        q2straj = sqza.view(pd.DataFrame)
        q2ftraj = fqza.view(pd.DataFrame)
        q2tree = tree.view(TreeNode)
        q2table = ctable.view(Table)

        # Next, run gemelli outside of QIIME 2. We're gonna check that
        # everything matches up.
        # ...First, though, we need to write the contents of self.q2table to a
        # BIOM file, so gemelli can understand it.
        # Derived from a line in test_standalone_ctf()
        # Run gemelli outside of QIIME 2...
        runner = CliRunner()
        result = runner.invoke(standalone_phylogenetic_ctf,
                               ['--in-biom',
                                self.in_table,
                                '--in-phylogeny',
                                self.in_tree,
                                '--sample-metadata-file',
                                self.in_meta,
                                '--individual-id-column',
                                'host_subject_id',
                                '--state-column-1',
                                'context',
                                '--output-dir',
                                self.out_])
        # check that exit code was 0 (indicating success)
        try:
            self.assertEqual(0, result.exit_code)
        except AssertionError:
            ex = result.exception
            error = Exception('Command failed with non-zero exit code')
            raise error.with_traceback(ex.__traceback__)
        # ...and read in the resulting output files. This code was derived from
        # test_standalone_ctf() elsewhere in gemelli's codebase.
        samp_res = read_csv(
            get_data_path('context-subject-ordination.tsv'),
            sep='\t',
            index_col=0)
        feat_res = read_csv(
            get_data_path('context-features-ordination.tsv'),
            sep='\t',
            index_col=0)
        bt_res = get_data_path('phylogenetic-table.biom')
        bt_res = load_table(bt_res)
        tree_res = get_data_path('labeled-phylogeny.nwk')
        tree_res = TreeNode.read(tree_res,
                                 format='newick')
        # Check that the trajectory matrix matches our expectations
        comp_col = ['PC1', 'PC2', 'PC3']
        assert_allclose(absolute_sort(samp_res[comp_col].values),
                        absolute_sort(q2straj[comp_col].values),
                        atol=.5)
        assert_allclose(absolute_sort(feat_res[comp_col].values),
                        absolute_sort(q2ftraj[comp_col].values),
                        atol=.5)
        assert_array_almost_equal(bt_res.matrix_data.toarray(),
                                  q2table.matrix_data.toarray())
        # check renamed names are consistent
        name_check_ = [x.name == y.name for x, y in zip(tree_res.postorder(),
                                                        q2tree.postorder())]
        name_check_ = all(name_check_)
        self.assertEqual(name_check_, True)

    def test_ctf_rank2(self):
        """Tests that ctf with rank < 3
        """
        # Run gemelli
        res = q2gemelli.actions.ctf(table=self.q2table,
                                    sample_metadata=self.q2meta,
                                    individual_id_column=self.subj,
                                    state_column=self.state,
                                    n_components=2)
        # check exit code was 0 (indicating success)
        self.assertEqual(len(res), 5)


if __name__ == "__main__":
    unittest.main()



import unittest
import pandas as pd
from skbio import TreeNode
from biom import load_table
from os.path import sep as os_path_sep
from click.testing import CliRunner
from skbio import OrdinationResults
from skbio.util import get_data_path
from numpy.testing import assert_array_almost_equal
from gemelli.scripts.__init__ import cli as sdc
from gemelli.testing import assert_ordinationresults_equal, CliTestCase


class Test_standalone_rpca(unittest.TestCase):

    def setUp(self):
        pass

    def test_standalone_rpca_rank_est(self):
        """Checks the standalone RPCA rank estimate
           is used instead of a explicit rank
           setting.
        """
        in_ = get_data_path('test.biom', subfolder='rpca_data')
        out_ = os_path_sep.join(in_.split(os_path_sep)[:-1])
        runner = CliRunner()
        result = runner.invoke(sdc.commands['auto-rpca'],
                               ['--in-biom', in_,
                                '--output-dir', out_])
        # Read the results
        dist_res = pd.read_csv(get_data_path('distance-matrix.tsv',
                                             subfolder='rpca_data'), sep='\t',
                               index_col=0)
        ord_res = OrdinationResults.read(get_data_path('ordination.txt',
                                                       subfolder='rpca_data'))

        # Read the expected results
        file_ = 'expected-est-distance-matrix.tsv'
        dist_exp = pd.read_csv(get_data_path(file_, subfolder='rpca_data'),
                               sep='\t', index_col=0)
        ord_exp = OrdinationResults.read(get_data_path(
                                         'expected-est-ordination.txt',
                                         subfolder='rpca_data'))

        # Check that the distance matrix matches our expectations
        assert_array_almost_equal(dist_res.values, dist_exp.values)

        # Check that the ordination results match our expectations -- checking
        # each value for both features and samples
        assert_ordinationresults_equal(ord_res, ord_exp)

        # check that exit code was 0 (indicating success)
        try:
            self.assertEqual(0, result.exit_code)
        except AssertionError:
            ex = result.exception
            error = Exception('Command failed with non-zero exit code')
            raise error.with_traceback(ex.__traceback__)

    def test_standalone_phylogenetic_rpca(self):
        """Checks the output gemelli's phylogenetic RPCA standalone script.

           This is more of an "integration test" than a unit test -- the
           details of the algorithm used by the standalone phylogenetic RPCA
           script are checked in more detail in gemelli/tests/test_optspace.py,
           etc.
        """
        in_ = get_data_path('test.biom', subfolder='rpca_data')
        in_tree_ = get_data_path('tree.nwk', subfolder='rpca_data')
        out_ = os_path_sep.join(in_.split(os_path_sep)[:-1])
        runner = CliRunner()
        result = runner.invoke(sdc.commands['phylogenetic-rpca'],
                               ['--in-biom', in_,
                                '--in-phylogeny', in_tree_,
                                '--output-dir', out_])
        # Read the results
        dist_res = pd.read_csv(get_data_path('distance-matrix.tsv',
                                             subfolder='rpca_data'),
                               sep='\t',
                               index_col=0)
        ord_res = OrdinationResults.read(get_data_path('ordination.txt',
                                                       subfolder='rpca_data'))
        tree_res = TreeNode.read(get_data_path('labeled-phylogeny.nwk',
                                               subfolder='rpca_data'),
                                 format='newick')
        bt_res = load_table(get_data_path('phylo-table.biom',
                                          subfolder='rpca_data'))

        # Read the expected results
        dist_exp = get_data_path('expected-phylo-distance-matrix.tsv',
                                 subfolder='rpca_data')
        dist_exp = pd.read_csv(dist_exp, sep='\t', index_col=0)
        ord_exp = get_data_path('expected-phylo-ordination.txt',
                                subfolder='rpca_data')
        ord_exp = OrdinationResults.read(ord_exp)
        tree_exp = get_data_path('expected-labeled-phylogeny.nwk',
                                 subfolder='rpca_data')
        tree_exp = TreeNode.read(tree_exp,
                                 format='newick')
        bt_exp = load_table(get_data_path('phylo-table.biom',
                            subfolder='rpca_data'))

        # check table values match
        assert_array_almost_equal(bt_res.matrix_data.toarray(),
                                  bt_exp.matrix_data.toarray())

        # check renamed names are consistent
        name_check_ = [x.name == y.name for x, y in zip(tree_res.postorder(),
                                                        tree_exp.postorder())]
        name_check_ = all(name_check_)
        self.assertEqual(name_check_, True)

        # Check that the distance matrix matches our expectations
        assert_array_almost_equal(dist_res.values, dist_exp.values)

        # Check that the ordination results match our expectations -- checking
        # each value for both features and samples
        assert_ordinationresults_equal(ord_res, ord_exp)

        # check that exit code was 0 (indicating success)
        try:
            self.assertEqual(0, result.exit_code)
        except AssertionError:
            ex = result.exception
            error = Exception('Command failed with non-zero exit code')
            raise error.with_traceback(ex.__traceback__)

    def test_standalone_rpca(self):
        """Checks the output produced by gemelli's RPCA standalone script.

           This is more of an "integration test" than a unit test -- the
           details of the algorithm used by the standalone RPCA script are
           checked in more detail in gemelli/tests/test_optspace.py, etc.
        """
        in_ = get_data_path('test.biom', subfolder='rpca_data')
        out_ = os_path_sep.join(in_.split(os_path_sep)[:-1])
        runner = CliRunner()
        result = runner.invoke(sdc.commands['rpca'],
                               ['--in-biom', in_,
                                '--output-dir', out_])
        # Read the results
        dist_res = pd.read_csv(get_data_path('distance-matrix.tsv',
                                             subfolder='rpca_data'),
                               sep='\t',
                               index_col=0)
        ord_res = OrdinationResults.read(get_data_path('ordination.txt',
                                                       subfolder='rpca_data'))

        # Read the expected results
        dist_exp = pd.read_csv(get_data_path('expected-distance-matrix.tsv',
                                             subfolder='rpca_data'),
                               sep='\t', index_col=0)
        ord_exp = OrdinationResults.read(get_data_path(
                                         'expected-ordination.txt',
                                         subfolder='rpca_data'))

        # Check that the distance matrix matches our expectations
        assert_array_almost_equal(dist_res.values, dist_exp.values)

        # Check that the ordination results match our expectations -- checking
        # each value for both features and samples
        assert_ordinationresults_equal(ord_res, ord_exp)

        # check that exit code was 0 (indicating success)
        try:
            self.assertEqual(0, result.exit_code)
        except AssertionError:
            ex = result.exception
            error = Exception('Command failed with non-zero exit code')
            raise error.with_traceback(ex.__traceback__)

    def test_standalone_rpca_n_components(self):
        """Tests the standalone RPCA script when n_components is 2
        """
        in_ = get_data_path('test.biom', subfolder='rpca_data')
        out_ = os_path_sep.join(in_.split(os_path_sep)[:-1])
        runner = CliRunner()
        # run the same command but with rank==2
        result = runner.invoke(sdc.commands['rpca'],
                               ['--in-biom', in_,
                                '--output-dir', out_,
                                '--n-components', 2,
                                '--max-iterations', 5])
        CliTestCase().assertExitCode(0, result)
        ord_res = OrdinationResults.read(get_data_path('ordination.txt',
                                                       subfolder='rpca_data'))
        # check that exit code was 0 (indicating success)
        try:
            self.assertEqual(0, result.exit_code)
        except AssertionError:
            ex = result.exception
            error = Exception('Command failed with non-zero exit code')
            raise error.with_traceback(ex.__traceback__)
        # check it contains three axis
        if len(ord_res.proportion_explained) == 3:
            pass


if __name__ == "__main__":
    unittest.main()



import os
import click
from .__init__ import cli
import pandas as pd
from biom.util import biom_open
from biom import load_table
from gemelli.ctf import (ctf_helper, phylogenetic_ctf_helper)
from gemelli.preprocessing import TaxonomyError
from gemelli._defaults import (DEFAULT_COMP, DEFAULT_MSC,
                               DEFAULT_MFC, DEFAULT_MTD,
                               DESC_MSC, DESC_MFC,
                               DEFAULT_TENSALS_MAXITER,
                               DEFAULT_FMETA, DESC_COMP,
                               DESC_OUT, DESC_ITERATIONSALS,
                               DESC_FMETA, DESC_COUNTS, DESC_SMETA,
                               DESC_SUBJ, DESC_COND, DESC_INIT,
                               DESC_MINDEPTH, DESC_TREE,
                               DEFAULT_MFF, DESC_MFF,
                               DESC_ITERATIONSRTPM, DEFAULT_COND)


@cli.command(name='phylogenetic-ctf')
@click.option('--in-biom',
              required=True,
              help=DESC_COUNTS)
@click.option('--in-phylogeny',
              help=DESC_TREE,
              required=True)
@click.option('--sample-metadata-file',
              required=True,
              help=DESC_SMETA)
@click.option('--individual-id-column',
              required=True,
              help=DESC_SUBJ)
@click.option('--state-column-1',
              required=True,
              help=DESC_COND)
@click.option('--output-dir',
              required=True,
              help=DESC_OUT)
@click.option('--n-components',
              default=DEFAULT_COMP,
              show_default=True,
              help=DESC_COMP)
@click.option('--min-sample-count',
              default=DEFAULT_MSC,
              show_default=True,
              help=DESC_MSC)
@click.option('--min-feature-count',
              default=DEFAULT_MFC,
              show_default=True,
              help=DESC_MFC)
@click.option('--min-feature-frequency',
              default=DEFAULT_MFF,
              show_default=True,
              help=DESC_MFF)
@click.option('--min-depth',
              default=DEFAULT_MTD,
              show_default=True,
              help=DESC_MINDEPTH)
@click.option('--max-iterations-als',
              default=DEFAULT_TENSALS_MAXITER,
              show_default=True,
              help=DESC_ITERATIONSALS)
@click.option('--max-iterations-rptm',
              default=DEFAULT_TENSALS_MAXITER,
              show_default=True,
              help=DESC_ITERATIONSRTPM)
@click.option('--n-initializations',
              default=DEFAULT_TENSALS_MAXITER,
              show_default=True,
              help=DESC_INIT)
@click.option('--feature-metadata-file',
              default=DEFAULT_FMETA,
              show_default=True,
              help=DESC_FMETA)
@click.option('--state-column-2',
              required=False,
              default=DEFAULT_COND,
              help=DESC_COND)
@click.option('--state-column-3',
              required=False,
              default=DEFAULT_COND,
              help=DESC_COND)
@click.option('--state-column-4',
              required=False,
              default=DEFAULT_COND,
              help=DESC_COND)
def standalone_phylogenetic_ctf(in_biom: str,
                                in_phylogeny: str,
                                sample_metadata_file: str,
                                individual_id_column: str,
                                state_column_1: str,
                                output_dir: str,
                                n_components: int,
                                min_sample_count: int,
                                min_feature_count: int,
                                min_feature_frequency: float,
                                min_depth: int,
                                max_iterations_als: int,
                                max_iterations_rptm: int,
                                n_initializations: int,
                                feature_metadata_file: None,
                                state_column_2: str,
                                state_column_3: str,
                                state_column_4: str) -> None:
    """Runs phylogenetically informed CTF with an rclr preprocessing step."""

    # generate state lists
    state_columns = [state for state in [state_column_1, state_column_2,
                                         state_column_3, state_column_4]
                     if state is not None]
    # import table
    table = load_table(in_biom)
    # import sample metadata
    sample_metadata = pd.read_csv(sample_metadata_file,
                                  sep='\t', index_col=0,
                                  low_memory=False)
    # import feature metadata if available
    feature_metadata = None
    if feature_metadata_file is not None:
        feature_metadata = pd.read_csv(feature_metadata_file,
                                       sep='\t', index_col=0,
                                       low_memory=False)
        try:
            taxonomy_table.set_index('Feature ID', inplace=True)
        except KeyError:
            raise TaxonomyError(
                        "Taxonomy file must have a column labled 'Feature ID'."
                        )
    # run CTF
    res_ = phylogenetic_ctf_helper(table,
                                   in_phylogeny,
                                   sample_metadata,
                                   individual_id_column,
                                   state_columns,
                                   n_components,
                                   min_sample_count,
                                   min_feature_count,
                                   min_feature_frequency,
                                   min_depth,
                                   max_iterations_als,
                                   max_iterations_rptm,
                                   n_initializations,
                                   feature_metadata)
    (state_ordn, ord_res, dists, straj, ftraj,
     phylogeny, counts_by_node,
     result_taxonomy, subject_table) = res_
    # If it doesn't already exist, create the output directory.
    # Note that there is technically a race condition here: it's ostensibly
    # possible that some process could delete the output directory after we
    # check that it exists here but before we write the output files to it.
    # However, in this case, we'd just get an error from skbio.io.util.open()
    # (which is called by skbio.OrdinationResults.write()), which makes sense.
    os.makedirs(output_dir, exist_ok=True)
    # write files to output directory
    # Note that this will overwrite files in the output directory that share
    # these filenames (analogous to QIIME 2's behavior if you specify the
    # --o-biplot and --o-distance-matrix options, but differing from QIIME 2's
    # behavior if you specify --output-dir instead).
    if result_taxonomy is not None:
        result_taxonomy.to_csv(os.path.join(output_dir, 't2t-taxonomy.tsv'),
                           sep='\t')
    ord_res.write(os.path.join(output_dir, 'ordination.txt'))
    # write each distance matrix
    for condition, dist in dists.items():
        dist.write(os.path.join(output_dir,
                                '%s-distance-matrix.tsv' % (str(condition))))
    # write each state ord
    for condition, ord_ in state_ordn.items():
        ord_.write(os.path.join(output_dir, '%s-ordination.txt' % (condition)))
    # write each trajectory
    for condition, traj in straj.items():
        traj.to_csv(
            os.path.join(
                output_dir,
                '%s-subject-ordination.tsv' %
                (str(condition))),
            sep='\t')
    # write each trajectory
    for condition, traj in ftraj.items():
        traj.to_csv(
            os.path.join(
                output_dir,
                '%s-features-ordination.tsv' %
                (str(condition))),
            sep='\t')
    # export phylogeny
    phylogeny.write(os.path.join(output_dir, 'labeled-phylogeny.nwk'))
    # write the vectorized count table for Qurro / log-ratios
    table_out_path = os.path.join(output_dir,
                                  'phylogenetic-table.biom')
    with biom_open(table_out_path, 'w') as f:
        counts_by_node.to_hdf5(f, "phylogenetic-ctf-count-table")
    table_out_path = os.path.join(output_dir,
                                  'subject-phylogenetic-table.biom')
    with biom_open(table_out_path, 'w') as f:
        subject_table.to_hdf5(f, "subject-ctf-count-table")


@cli.command(name='ctf')
@click.option(
    '--in-biom',
    required=True,
    help=DESC_COUNTS)
@click.option(
    '--sample-metadata-file',
    required=True,
    help=DESC_SMETA)
@click.option(
    '--individual-id-column',
    required=True,
    help=DESC_SUBJ)
@click.option(
    '--state-column-1',
    required=True,
    help=DESC_COND)
@click.option(
    '--output-dir',
    required=True,
    help=DESC_OUT)
@click.option(
    '--n-components',
    default=DEFAULT_COMP,
    show_default=True,
    help=DESC_COMP)
@click.option(
    '--min-sample-count',
    default=DEFAULT_MSC,
    show_default=True,
    help=DESC_MSC)
@click.option(
    '--min-feature-count',
    default=DEFAULT_MFC,
    show_default=True,
    help=DESC_MFC)
@click.option(
    '--min-feature-frequency',
    default=DEFAULT_MFF,
    show_default=True,
    help=DESC_MFF)
@click.option(
    '--max-iterations-als',
    default=DEFAULT_TENSALS_MAXITER,
    show_default=True,
    help=DESC_ITERATIONSALS)
@click.option(
    '--max-iterations-rptm',
    default=DEFAULT_TENSALS_MAXITER,
    show_default=True,
    help=DESC_ITERATIONSRTPM)
@click.option(
    '--n-initializations',
    default=DEFAULT_TENSALS_MAXITER,
    show_default=True,
    help=DESC_INIT)
@click.option(
    '--feature-metadata-file',
    default=DEFAULT_FMETA,
    show_default=True,
    help=DESC_FMETA)
@click.option(
    '--state-column-2',
    required=False,
    default=DEFAULT_COND,
    help=DESC_COND)
@click.option(
    '--state-column-3',
    required=False,
    default=DEFAULT_COND,
    help=DESC_COND)
@click.option(
    '--state-column-4',
    required=False,
    default=DEFAULT_COND,
    help=DESC_COND)
def standalone_ctf(in_biom: str,
                   sample_metadata_file: str,
                   individual_id_column: str,
                   state_column_1: str,
                   output_dir: str,
                   n_components: int,
                   min_sample_count: int,
                   min_feature_count: int,
                   min_feature_frequency: float,
                   max_iterations_als: int,
                   max_iterations_rptm: int,
                   n_initializations: int,
                   feature_metadata_file: str,
                   state_column_2: str,
                   state_column_3: str,
                   state_column_4: str) -> None:
    """Runs CTF with an rclr preprocessing step."""

    # generate state lists
    state_columns = [state for state in [state_column_1, state_column_2,
                                         state_column_3, state_column_4]
                     if state is not None]
    # import table
    table = load_table(in_biom)
    # import sample metadata
    sample_metadata = pd.read_csv(sample_metadata_file,
                                  sep='\t', index_col=0,
                                  low_memory=False)
    # import feature metadata if available
    if feature_metadata_file is not None:
        feature_metadata = pd.read_csv(feature_metadata_file,
                                       sep='\t', index_col=0,
                                       low_memory=False)
    else:
        feature_metadata = None
    # run CTF
    res_ = ctf_helper(table,
                      sample_metadata,
                      individual_id_column,
                      state_columns,
                      n_components,
                      min_sample_count,
                      min_feature_count,
                      min_feature_frequency,
                      max_iterations_als,
                      max_iterations_rptm,
                      n_initializations,
                      feature_metadata)
    state_ordn, ord_res, dists, straj, ftraj = res_
    # If it doesn't already exist, create the output directory.
    # Note that there is technically a race condition here: it's ostensibly
    # possible that some process could delete the output directory after we
    # check that it exists here but before we write the output files to it.
    # However, in this case, we'd just get an error from skbio.io.util.open()
    # (which is called by skbio.OrdinationResults.write()), which makes sense.
    os.makedirs(output_dir, exist_ok=True)
    # write files to output directory
    # Note that this will overwrite files in the output directory that share
    # these filenames (analogous to QIIME 2's behavior if you specify the
    # --o-biplot and --o-distance-matrix options, but differing from QIIME 2's
    # behavior if you specify --output-dir instead).
    ord_res.write(os.path.join(output_dir, 'ordination.txt'))
    # write each distance matrix
    for condition, dist in dists.items():
        dist.write(os.path.join(output_dir,
                                '%s-distance-matrix.tsv' % (str(condition))))
    # write each state ord
    for condition, ord_ in state_ordn.items():
        ord_.write(os.path.join(output_dir, '%s-ordination.txt' % (condition)))
    # write each trajectory
    for condition, traj in straj.items():
        traj.to_csv(
            os.path.join(
                output_dir,
                '%s-subject-ordination.tsv' %
                (str(condition))),
            sep='\t')
    # write each trajectory
    for condition, traj in ftraj.items():
        traj.to_csv(
            os.path.join(
                output_dir,
                '%s-features-ordination.tsv' %
                (str(condition))),
            sep='\t')



# ----------------------------------------------------------------------------
# Copyright (c) 2019--, gemelli development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# ----------------------------------------------------------------------------

import qiime2.plugin
import qiime2.sdk
import importlib
from gemelli import __version__
from gemelli.ctf import (ctf, phylogenetic_ctf,
                         phylogenetic_ctf_without_taxonomy,
                         phylogenetic_ctf_with_taxonomy)
from gemelli.rpca import (rpca, auto_rpca,
                          phylogenetic_rpca_with_taxonomy,
                          phylogenetic_rpca_without_taxonomy)
from gemelli.preprocessing import (rclr_transformation,
                                   phylogenetic_rclr_transformation)
from ._type import SampleTrajectory, FeatureTrajectory
from ._format import TrajectoryDirectoryFormat
from qiime2.plugin import (Properties, Int, Float, Metadata, Str)
from q2_types.ordination import PCoAResults
from q2_types.distance_matrix import DistanceMatrix
from q2_types.sample_data import SampleData
from q2_types.feature_data import FeatureData
from q2_types.tree import Phylogeny, Rooted
from q2_types.feature_table import FeatureTable, Frequency, Composition
from q2_types.feature_data import Taxonomy
from qiime2.plugin import Metadata
from gemelli._defaults import (DESC_COMP, DESC_ITERATIONSALS,
                               DESC_BIN, DESC_SMETA, DESC_TREE,
                               DESC_SUBJ, DESC_COND, DESC_INIT,
                               DESC_ITERATIONSRTPM, DESC_MINDEPTH,
                               QLOAD, QDIST, QORD, QSOAD, QRCLR,
                               DESC_MSC, DESC_MFC, QBIPLOT,
                               QTREE, QTREECOUNT, QADIST,
                               DESC_ITERATIONS, DESC_MFF, DESC_TAX_Q2,
                               DESC_T2T_TAX, DESC_STBL)

citations = qiime2.plugin.Citations.load(
    'citations.bib', package='gemelli')

plugin = qiime2.plugin.Plugin(
    name='gemelli',
    version=__version__,
    website="https://github.com/biocore/gemelli",
    citations=[citations['Martino2019'],
               citations['Martino2020']],
    short_description=('Plugin for Compositional Tensor Factorization'),
    description=('This is a QIIME 2 plugin supporting Robust Aitchison on '
                 'feature tables'),
    package='gemelli')

plugin.methods.register_function(
    function=rclr_transformation,
    inputs={'table': FeatureTable[Frequency]},
    parameters=None,
    outputs=[('rclr_table', FeatureTable[Composition])],
    input_descriptions={'table': DESC_BIN},
    parameter_descriptions=None,
    output_descriptions={'rclr_table': QRCLR},
    name=('Robust centered log-ratio (rclr) transformation.'
          'Note: This is run automatically '
          'within CTF/RPCA/Auto-RPCA so there no '
          'need to run rclr before those functions.'),
    description=("A robust centered log-ratio transformation of only "
                 "the observed values (non-zero) of the input table."),
    citations=[citations['Martino2019']]
)

plugin.methods.register_function(
    function=phylogenetic_rclr_transformation,
    inputs={'table': FeatureTable[Frequency],
            'phylogeny': Phylogeny[Rooted]},
    parameters={'min_depth': Int},
    outputs=[('counts_by_node', FeatureTable[Frequency]),
             ('rclr_table', FeatureTable[Composition]),
             ('counts_by_node_tree', Phylogeny[Rooted])],
    input_descriptions={'table': DESC_BIN,
                        'phylogeny': DESC_TREE},
    parameter_descriptions={'min_depth': DESC_MINDEPTH},
    output_descriptions={'counts_by_node': QTREECOUNT,
                         'rclr_table': QRCLR,
                         'counts_by_node_tree': QTREE},
    name=('Phylogenetic Robust centered log-ratio (rclr) transformation.'
          'Note: This is run automatically '
          'within phylogenetic-CTF/RPCA so there no '
          'need to run rclr before those functions.'),
    description=("A phylogenetic robust centered log-ratio transformation "
                 "of only the observed values (non-zero) of the input table."),
    citations=[citations['Martino2019']]
)

plugin.methods.register_function(
    function=ctf,
    inputs={'table': FeatureTable[Frequency]},
    parameters={'sample_metadata': Metadata,
                'individual_id_column': Str,
                'state_column': Str,
                'n_components': Int,
                'min_sample_count': Int,
                'min_feature_count': Int,
                'min_feature_frequency': Float,
                'max_iterations_als': Int,
                'max_iterations_rptm': Int,
                'n_initializations': Int,
                'feature_metadata': Metadata},
    outputs=[('subject_biplot', PCoAResults % Properties("biplot")),
             ('state_biplot', PCoAResults % Properties("biplot")),
             ('distance_matrix', DistanceMatrix),
             ('state_subject_ordination', SampleData[SampleTrajectory]),
             ('state_feature_ordination', FeatureData[FeatureTrajectory])],
    input_descriptions={'table': DESC_BIN},
    parameter_descriptions={'sample_metadata': DESC_SMETA,
                            'individual_id_column': DESC_SUBJ,
                            'state_column': DESC_COND,
                            'n_components': DESC_COMP,
                            'min_sample_count': DESC_MSC,
                            'min_feature_count': DESC_MFC,
                            'min_feature_frequency': DESC_MFF,
                            'max_iterations_als': DESC_ITERATIONSALS,
                            'max_iterations_rptm': DESC_ITERATIONSRTPM,
                            'n_initializations': DESC_INIT},
    output_descriptions={'subject_biplot': QLOAD,
                         'state_biplot': QSOAD,
                         'distance_matrix': QDIST,
                         'state_subject_ordination': QORD,
                         'state_feature_ordination': QORD},
    name='Compositional Tensor Factorization (CTF) with mode 3 tensor. This '
         'means subjects have repeated measures across only one '
         'axis (e.g. time or space).',
    description=("Gemelli resolves spatiotemporal subject variation and the"
                 " biological features that separate them. In this case, a "
                 "subject may have several paired samples, where each sample"
                 " may be a time point. The output is akin to conventional "
                 "beta-diversity analyses but with the paired component "
                 "integrated in the dimensionality reduction."),
    citations=[citations['Martino2020']]
)

plugin.methods.register_function(
    function=phylogenetic_ctf,
    inputs={'table': FeatureTable[Frequency],
            'phylogeny': Phylogeny[Rooted]},
    parameters={'sample_metadata': Metadata,
                'individual_id_column': Str,
                'state_column': Str,
                'taxonomy': Metadata,
                'n_components': Int,
                'min_sample_count': Int,
                'min_feature_count': Int,
                'min_feature_frequency': Float,
                'min_depth': Int,
                'max_iterations_als': Int,
                'max_iterations_rptm': Int,
                'n_initializations': Int},
    outputs=[('subject_biplot', PCoAResults % Properties("biplot")),
             ('state_biplot', PCoAResults % Properties("biplot")),
             ('distance_matrix', DistanceMatrix),
             ('state_subject_ordination', SampleData[SampleTrajectory]),
             ('state_feature_ordination', FeatureData[FeatureTrajectory]),
             ('counts_by_node_tree', Phylogeny[Rooted]),
             ('counts_by_node', FeatureTable[Frequency]),
             ('t2t_taxonomy', FeatureData[Taxonomy]),
             ('subject_table', FeatureTable[Frequency])],
    input_descriptions={'table': DESC_BIN,
                        'phylogeny': DESC_TREE},
    parameter_descriptions={'sample_metadata': DESC_SMETA,
                            'individual_id_column': DESC_SUBJ,
                            'state_column': DESC_COND,
                            'taxonomy': DESC_TAX_Q2,
                            'n_components': DESC_COMP,
                            'min_sample_count': DESC_MSC,
                            'min_feature_count': DESC_MFC,
                            'min_feature_frequency': DESC_MFF,
                            'max_iterations_als': DESC_ITERATIONSALS,
                            'max_iterations_rptm': DESC_ITERATIONSRTPM,
                            'n_initializations': DESC_INIT},
    output_descriptions={'subject_biplot': QLOAD,
                         'state_biplot': QSOAD,
                         'distance_matrix': QDIST,
                         'state_subject_ordination': QORD,
                         'state_feature_ordination': QORD,
                         'counts_by_node_tree': QTREE,
                         'counts_by_node': QTREECOUNT,
                         't2t_taxonomy': DESC_T2T_TAX,
                         'subject_table': DESC_STBL},
    name='Phylogenetic Compositional Tensor Factorization (CTF) '
         'with mode 3 tensor. This means subjects have repeated '
         'measures across only one axis (e.g. time or space). '
         'The input taxonomy is used to produce a new taxonomy '
         'label for each node in the tree based on the lowest '
         'common anscestor. Note: equivelent to '
         'phylogenetic-ctf-with-taxonomy',
    description=("Gemelli resolves spatiotemporal subject variation and the"
                 " biological features that separate them. In this case, a "
                 "subject may have several paired samples, where each sample"
                 " may be a time point. The output is akin to conventional "
                 "beta-diversity analyses but with the paired component "
                 "integrated in the dimensionality reduction."),
    citations=[citations['Martino2020']]
)

plugin.methods.register_function(
    function=phylogenetic_ctf_with_taxonomy,
    inputs={'table': FeatureTable[Frequency],
            'phylogeny': Phylogeny[Rooted]},
    parameters={'sample_metadata': Metadata,
                'taxonomy': Metadata,
                'individual_id_column': Str,
                'state_column': Str,
                'n_components': Int,
                'min_sample_count': Int,
                'min_feature_count': Int,
                'min_feature_frequency': Float,
                'min_depth': Int,
                'max_iterations_als': Int,
                'max_iterations_rptm': Int,
                'n_initializations': Int},
    outputs=[('subject_biplot', PCoAResults % Properties("biplot")),
             ('state_biplot', PCoAResults % Properties("biplot")),
             ('distance_matrix', DistanceMatrix),
             ('state_subject_ordination', SampleData[SampleTrajectory]),
             ('state_feature_ordination', FeatureData[FeatureTrajectory]),
             ('counts_by_node_tree', Phylogeny[Rooted]),
             ('counts_by_node', FeatureTable[Frequency]),
             ('t2t_taxonomy', FeatureData[Taxonomy]),
             ('subject_table', FeatureTable[Frequency])],
    input_descriptions={'table': DESC_BIN,
                        'phylogeny': DESC_TREE},
    parameter_descriptions={'sample_metadata': DESC_SMETA,
                            'taxonomy': DESC_TAX_Q2,
                            'individual_id_column': DESC_SUBJ,
                            'state_column': DESC_COND,
                            'n_components': DESC_COMP,
                            'min_sample_count': DESC_MSC,
                            'min_feature_count': DESC_MFC,
                            'min_feature_frequency': DESC_MFF,
                            'max_iterations_als': DESC_ITERATIONSALS,
                            'max_iterations_rptm': DESC_ITERATIONSRTPM,
                            'n_initializations': DESC_INIT},
    output_descriptions={'subject_biplot': QLOAD,
                         'state_biplot': QSOAD,
                         'distance_matrix': QDIST,
                         'state_subject_ordination': QORD,
                         'state_feature_ordination': QORD,
                         'counts_by_node_tree': QTREE,
                         'counts_by_node': QTREECOUNT,
                         't2t_taxonomy': DESC_T2T_TAX,
                         'subject_table': DESC_STBL},
    name='Phylogenetic Compositional Tensor Factorization (CTF) '
         'with mode 3 tensor. This means subjects have repeated '
         'measures across only one axis (e.g. time or space). '
         'The input taxonomy is used to produce a new taxonomy '
         'label for each node in the tree based on the lowest '
         'common anscestor.',
    description=("Gemelli resolves spatiotemporal subject variation and the"
                 " biological features that separate them. In this case, a "
                 "subject may have several paired samples, where each sample"
                 " may be a time point. The output is akin to conventional "
                 "beta-diversity analyses but with the paired component "
                 "integrated in the dimensionality reduction."),
    citations=[citations['Martino2020']]
)

plugin.methods.register_function(
    function=phylogenetic_ctf_without_taxonomy,
    inputs={'table': FeatureTable[Frequency],
            'phylogeny': Phylogeny[Rooted]},
    parameters={'sample_metadata': Metadata,
                'individual_id_column': Str,
                'state_column': Str,
                'n_components': Int,
                'min_sample_count': Int,
                'min_feature_count': Int,
                'min_feature_frequency': Float,
                'min_depth': Int,
                'max_iterations_als': Int,
                'max_iterations_rptm': Int,
                'n_initializations': Int},
    outputs=[('subject_biplot', PCoAResults % Properties("biplot")),
             ('state_biplot', PCoAResults % Properties("biplot")),
             ('distance_matrix', DistanceMatrix),
             ('state_subject_ordination', SampleData[SampleTrajectory]),
             ('state_feature_ordination', FeatureData[FeatureTrajectory]),
             ('counts_by_node_tree', Phylogeny[Rooted]),
             ('counts_by_node', FeatureTable[Frequency]),
             ('subject_table', FeatureTable[Frequency])],
    input_descriptions={'table': DESC_BIN,
                        'phylogeny': DESC_TREE},
    parameter_descriptions={'sample_metadata': DESC_SMETA,
                            'individual_id_column': DESC_SUBJ,
                            'state_column': DESC_COND,
                            'n_components': DESC_COMP,
                            'min_sample_count': DESC_MSC,
                            'min_feature_count': DESC_MFC,
                            'min_feature_frequency': DESC_MFF,
                            'max_iterations_als': DESC_ITERATIONSALS,
                            'max_iterations_rptm': DESC_ITERATIONSRTPM,
                            'n_initializations': DESC_INIT},
    output_descriptions={'subject_biplot': QLOAD,
                         'state_biplot': QSOAD,
                         'distance_matrix': QDIST,
                         'state_subject_ordination': QORD,
                         'state_feature_ordination': QORD,
                         'counts_by_node_tree': QTREE,
                         'counts_by_node': QTREECOUNT,
                         'subject_table' : DESC_STBL},
    name='Phylogenetic Compositional Tensor Factorization (CTF) '
         'with mode 3 tensor. This means subjects have repeated '
         'measures across only one axis (e.g. time or space). '
         'Note: This does not require/output a taxonomy. '
         'A taxonomy for the input phylogeny will still  be valid '
         'for tip level features however, '
         'if taxonomy is required for internal features please use '
         'phylogenetic-ctf-with-taxonomy.',
    description=("Gemelli resolves spatiotemporal subject variation and the"
                 " biological features that separate them. In this case, a "
                 "subject may have several paired samples, where each sample"
                 " may be a time point. The output is akin to conventional "
                 "beta-diversity analyses but with the paired component "
                 "integrated in the dimensionality reduction."),
    citations=[citations['Martino2020']]
)

plugin.methods.register_function(
    function=rpca,
    inputs={'table': FeatureTable[Frequency]},
    parameters={'n_components': Int,
                'min_sample_count': Int,
                'min_feature_count': Int,
                'min_feature_frequency': Float,
                'max_iterations': Int},
    outputs=[('biplot', PCoAResults % Properties("biplot")),
             ('distance_matrix', DistanceMatrix)],
    input_descriptions={'table': DESC_BIN},
    parameter_descriptions={'n_components': DESC_COMP,
                            'min_sample_count': DESC_MSC,
                            'min_feature_count': DESC_MFC,
                            'min_feature_frequency': DESC_MFF,
                            'max_iterations': DESC_ITERATIONS},
    output_descriptions={'biplot': QBIPLOT,
                         'distance_matrix': QADIST},
    name='(Robust Aitchison) RPCA with manually chosen n_components.',
    description=("Performs robust center log-ratio transform "
                 "robust PCA and ranks the features by the "
                 "loadings of the resulting SVD."),
    citations=[citations['Martino2019']]
)

plugin.methods.register_function(
    function=phylogenetic_rpca_with_taxonomy,
    inputs={'table': FeatureTable[Frequency],
            'phylogeny': Phylogeny[Rooted],
            },
    parameters={
        'taxonomy': Metadata,
        'n_components': Int,
        'min_sample_count': Int,
        'min_feature_count': Int,
        'min_feature_frequency': Float,
        'min_depth': Int,
        'max_iterations': Int},
    outputs=[
        ('biplot', PCoAResults % Properties("biplot")),
        ('distance_matrix', DistanceMatrix),
        ('counts_by_node_tree', Phylogeny[Rooted]),
        ('counts_by_node', FeatureTable[Frequency]),
        ('t2t_taxonomy', FeatureData[Taxonomy])],
    input_descriptions={'table': DESC_BIN, 'phylogeny': DESC_TREE},
    parameter_descriptions={'taxonomy': DESC_TAX_Q2,
                            'n_components': DESC_COMP,
                            'min_sample_count': DESC_MSC,
                            'min_feature_count': DESC_MFC,
                            'min_feature_frequency': DESC_MFF,
                            'min_depth': DESC_MINDEPTH,
                            'max_iterations': DESC_ITERATIONS},
    output_descriptions={
        'biplot': QBIPLOT,
        'distance_matrix': QADIST,
        'counts_by_node_tree': QTREE,
        'counts_by_node': QTREECOUNT,
        't2t_taxonomy': DESC_T2T_TAX},
    name='Phylogenetic (Robust Aitchison) RPCA.',
    description=("Performs phylogenetic robust center log-ratio transform "
                 "robust PCA and ranks the features by the "
                 "loadings of the resulting SVD"),
    citations=[citations['Martino2019']]
)

plugin.methods.register_function(
    function=phylogenetic_rpca_without_taxonomy,
    inputs={'table': FeatureTable[Frequency],
            'phylogeny': Phylogeny[Rooted]},
    parameters={
        'n_components': Int,
        'min_sample_count': Int,
        'min_feature_count': Int,
        'min_feature_frequency': Float,
        'min_depth': Int,
        'max_iterations': Int},
    outputs=[
        ('biplot', PCoAResults % Properties("biplot")),
        ('distance_matrix', DistanceMatrix),
        ('counts_by_node_tree', Phylogeny[Rooted]),
        ('counts_by_node', FeatureTable[Frequency])],
    input_descriptions={'table': DESC_BIN, 'phylogeny': DESC_TREE},
    parameter_descriptions={'n_components': DESC_COMP,
                            'min_sample_count': DESC_MSC,
                            'min_feature_count': DESC_MFC,
                            'min_feature_frequency': DESC_MFF,
                            'min_depth': DESC_MINDEPTH,
                            'max_iterations': DESC_ITERATIONS},
    output_descriptions={
        'biplot': QBIPLOT,
        'distance_matrix': QADIST,
        'counts_by_node_tree': QTREE,
        'counts_by_node': QTREECOUNT},
    name=('Phylogenetic (Robust Aitchison) RPCA. '
          'Note: This does not require/output a taxonomy. '
          'A taxonomy for the input phylogeny will still  be valid '
          'for tip level features however, '
          'if taxonomy is required for internal features please use '
          'phylogenetic-rpca-with-taxonomy.'),
    description=("Performs phylogenetic robust center log-ratio transform "
                 "robust PCA and ranks the features by the "
                 "loadings of the resulting SVD"),
    citations=[citations['Martino2019']]
)

plugin.methods.register_function(
    function=auto_rpca,
    inputs={'table': FeatureTable[Frequency]},
    parameters={'min_sample_count': Int,
                'min_feature_count': Int,
                'min_feature_frequency': Float,
                'max_iterations': Int},
    outputs=[('biplot', PCoAResults % Properties("biplot")),
             ('distance_matrix', DistanceMatrix)],
    input_descriptions={'table': DESC_BIN},
    parameter_descriptions={'min_sample_count': DESC_MSC,
                            'min_feature_count': DESC_MFC,
                            'min_feature_frequency': DESC_MFF,
                            'max_iterations': DESC_ITERATIONS},
    output_descriptions={'biplot': QBIPLOT,
                         'distance_matrix': QADIST},
    name='(Robust Aitchison) RPCA with n_components automatically detected.',
    description=("Performs robust center log-ratio transform "
                 "robust PCA and ranks the features by the "
                 "loadings of the resulting SVD. Automatically"
                 " estimates the underlying rank (i.e. n-components)."),
    citations=[citations['Martino2019']]
)

plugin.register_semantic_types(SampleTrajectory, FeatureTrajectory)
plugin.register_semantic_type_to_format(
    SampleData[SampleTrajectory],
    artifact_format=TrajectoryDirectoryFormat)
plugin.register_semantic_type_to_format(
    FeatureData[FeatureTrajectory],
    artifact_format=TrajectoryDirectoryFormat)
plugin.register_formats(TrajectoryDirectoryFormat)
importlib.import_module('gemelli.q2._transformer')



import click
from gemelli import __version__
from importlib import import_module


def _terribly_handle_brokenpipeerror():
    # based off http://stackoverflow.com/a/34299346
    import os
    import sys
    sys.stdout = os.fdopen(1, 'w')


@click.group()
@click.version_option(version=__version__)
@click.pass_context
def cli(ctx):
    ctx.call_on_close(_terribly_handle_brokenpipeerror)


import_module('gemelli.scripts._standalone_transforms')
import_module('gemelli.scripts._standalone_ctf')
import_module('gemelli.scripts._standalone_rpca')



import unittest
import os
import numpy as np
from skbio import TreeNode
import numpy.testing as npt
from biom import load_table, Table
from biom.util import biom_open
from os.path import sep as os_path_sep
from click.testing import CliRunner
from skbio.util import get_data_path
from gemelli.scripts.__init__ import cli as sdc
from gemelli.testing import CliTestCase
from numpy.testing import assert_array_almost_equal


class Test_standalone_rclr(unittest.TestCase):

    def setUp(self):
        self.cdata = np.array([[3, 3, 0],
                               [0, 4, 2]])
        self.true = np.array([[0.0, 0.0, np.nan],
                              [np.nan, 0.34657359, -0.34657359]])
        pass

    def test_standalone_rclr(self):
        """Test the standalone rlcr."""
        # make mock table to write
        samps_ids = ['s%i' % i for i in range(self.cdata.shape[0])]
        feats_ids = ['f%i' % i for i in range(self.cdata.shape[1])]
        table_test = Table(self.cdata.T, feats_ids, samps_ids)

        # write table
        in_ = get_data_path('test.biom', subfolder='rpca_data')
        out_path = os_path_sep.join(in_.split(os_path_sep)[:-1])
        test_path = os.path.join(out_path, 'rclr-test.biom')
        with biom_open(test_path, 'w') as wf:
            table_test.to_hdf5(wf, "test")

        # run rclr command
        runner = CliRunner()
        result = runner.invoke(sdc.commands['rclr'],
                               ['--in-biom', test_path,
                                '--output-dir', out_path])
        out_table = get_data_path('rclr-table.biom',
                                  subfolder='rpca_data')
        res_table = load_table(out_table)
        test_cmat = res_table.matrix_data.toarray().T
        npt.assert_allclose(test_cmat, self.true)
        # Lastly, check that exit code was 0 (indicating success)
        CliTestCase().assertExitCode(0, result)

    def test_standalone_phylogenetic_rclr(self):
        """Test the standalone phylogenetic-rlcr."""

        # write table
        in_table = get_data_path('test.biom', subfolder='rpca_data')
        in_tree = get_data_path('tree.nwk', subfolder='rpca_data')
        out_path = os_path_sep.join(in_table.split(os_path_sep)[:-1])

        # run phylogenetic-rclr
        runner = CliRunner()
        result = runner.invoke(sdc.commands['phylogenetic-rclr'],
                               ['--in-biom', in_table,
                                '--in-phylogeny', in_tree,
                                '--output-dir', out_path])

        # import expected
        tree_exp = get_data_path('expected-labeled-phylogeny.nwk',
                                 subfolder='rpca_data')
        tree_exp = TreeNode.read(tree_exp, format='newick')
        counts_exp = get_data_path('expected-phylogenetic-count-table.biom',
                                   subfolder='rpca_data')
        counts_exp = load_table(counts_exp)
        rclr_exp = get_data_path('expected-phylogenetic-rclr-table.biom',
                                 subfolder='rpca_data')
        rclr_exp = load_table(rclr_exp)

        # import result
        tree_res = get_data_path('labeled-phylogeny.nwk',
                                 subfolder='rpca_data')
        tree_res = TreeNode.read(tree_res, format='newick')
        counts_res = get_data_path('phylogenetic-count-table.biom',
                                   subfolder='rpca_data')
        counts_res = load_table(counts_res)
        rclr_res = get_data_path('phylogenetic-rclr-table.biom',
                                 subfolder='rpca_data')
        rclr_res = load_table(rclr_res)

        # check table values match
        assert_array_almost_equal(counts_exp.matrix_data.toarray(),
                                  counts_res.matrix_data.toarray())
        assert_array_almost_equal(rclr_exp.matrix_data.toarray(),
                                  rclr_res.matrix_data.toarray())

        # check renamed names are consistent
        name_check_ = [x.name == y.name for x, y in zip(tree_res.postorder(),
                                                        tree_exp.postorder())]
        name_check_ = all(name_check_)
        self.assertEqual(name_check_, True)

        # Lastly, check that exit code was 0 (indicating success)
        CliTestCase().assertExitCode(0, result)
